"""
Tests for the scaler factory classes and helper functions.
"""

from __future__ import annotations

from unittest.mock import MagicMock, Mock

import pytest

from dxtbx.model import Crystal
from libtbx import phil

from dials.algorithms.scaling.error_model.error_model import BasicErrorModel
from dials.algorithms.scaling.scaler import (
    MultiScaler,
    NullScaler,
    SingleScaler,
    TargetScaler,
)
from dials.algorithms.scaling.scaler_factory import (
    MultiScalerFactory,
    SingleScalerFactory,
    TargetScalerFactory,
    create_scaler,
)
from dials.array_family import flex
from dials.util.options import ArgumentParser


def generated_refl(not_integrated=False, idval=0):
    """Generate a test reflection table."""
    reflections = flex.reflection_table()
    reflections["intensity.prf.value"] = flex.double([1.0, 10.0, 100.0, 1.0])
    reflections["intensity.prf.variance"] = flex.double([1.0, 10.0, 100.0, 1.0])
    reflections["intensity.sum.value"] = flex.double([12.0, 120.0, 1200.0, 21.0])
    reflections["intensity.sum.variance"] = flex.double([12.0, 120.0, 2100.0, 1.0])
    # reflections['inverse_scale_factor'] = flex.double(4, 1.0)
    reflections["miller_index"] = flex.miller_index(
        [(1, 0, 0), (0, 0, 1), (2, 0, 0), (2, 2, 2)]
    )  # don't change
    reflections["d"] = flex.double([0.8, 2.0, 2.0, 0.0])  # don't change
    reflections["partiality"] = flex.double([1.0, 1.0, 1.0, 1.0])
    reflections["xyzobs.px.value"] = flex.vec3_double(
        [(0.0, 0.0, 0.0), (0.0, 0.0, 5.0), (0.0, 0.0, 10.0), (0.0, 0.0, 10.0)]
    )
    reflections["s1"] = flex.vec3_double(
        [(0.0, 0.1, 1.0), (0.0, 0.1, 1.0), (0.0, 0.1, 1.0), (0.0, 0.1, 1.0)]
    )
    if not_integrated:
        reflections.set_flags(
            flex.bool([False, False, False, False]), reflections.flags.integrated
        )
    else:
        reflections.set_flags(
            flex.bool([True, True, False, False]), reflections.flags.integrated
        )
    reflections.set_flags(
        flex.bool([False, False, True, True]), reflections.flags.bad_for_scaling
    )
    reflections["id"] = flex.int(4, idval)
    reflections.experiment_identifiers()[idval] = str(idval)
    return reflections


@pytest.fixture
def refl_to_filter():
    """Generate a separate reflection table for filtering"""
    reflections = flex.reflection_table()
    reflections["partiality"] = flex.double([0.1, 1.0, 1.0, 1.0, 1.0, 1.0])
    reflections["id"] = flex.int(reflections.size(), 0)
    reflections["intensity.sum.value"] = flex.double([1.0, 1.0, 1.0, 1.0, 1.0, 1.0])
    reflections["intensity.sum.variance"] = flex.double([1.0, 1.0, 1.0, 1.0, 1.0, 1.0])
    reflections.set_flags(
        flex.bool([True, False, True, True, True, True]),
        reflections.flags.integrated_sum,
    )
    return reflections


@pytest.fixture
def prf_sum_refl_to_filter():
    """Generate a separate reflection table for filtering"""
    reflections = flex.reflection_table()
    reflections["partiality"] = flex.double(5, 1.0)
    reflections["id"] = flex.int(reflections.size(), 0)
    reflections.experiment_identifiers()[0] = "0"
    reflections["intensity.sum.value"] = flex.double([1.0, 2.0, 3.0, 4.0, 5.0])
    reflections["intensity.sum.variance"] = flex.double(5, 1.0)
    reflections["intensity.prf.value"] = flex.double([11.0, 12.0, 13.0, 14.0, 15.0])
    reflections["intensity.prf.variance"] = flex.double(5, 1.0)
    reflections["miller_index"] = flex.miller_index([(0, 0, 1)] * 5)
    reflections.set_flags(
        flex.bool([False, False, True, True, True]),
        reflections.flags.integrated_sum,
    )
    reflections.set_flags(
        flex.bool([True, False, False, True, True]),
        reflections.flags.integrated_prf,
    )
    return reflections


def test_refl_and_exp(mock_scaling_component, idval=0):
    r = test_refl(idval=idval)
    exp = mock_exp(mock_scaling_component, idval=idval)
    return r, exp


def test_refl_and_exp_list(mock_scaling_component, n=1):
    rlist = []
    explist = []
    for i in range(n):
        rlist.append(test_refl(idval=i))
        explist.append(mock_exp(mock_scaling_component, idval=i))
    return rlist, explist


def test_refl(idval=0):
    """Generate a test reflection table."""
    return generated_refl(idval=idval)


@pytest.fixture
def refl_list():
    """Make a list of three reflection tables."""
    refl_list = [generated_refl()]
    refl_list.append(generated_refl())
    refl_list.append(generated_refl())
    return refl_list


@pytest.fixture
def generated_param():
    """Generate a param phil scope."""
    phil_scope = phil.parse(
        """
      include scope dials.algorithms.scaling.scaling_options.phil_scope
  """,
        process_includes=True,
    )
    parser = ArgumentParser(phil=phil_scope, check_format=False)
    parameters, _ = parser.parse_args(args=[], quick_parse=True, show_diff_phil=False)
    parameters.__inject__("model", "KB")
    parameters.scaling_options.free_set_percentage = 50.0
    parameters.scaling_options.emax = 0
    parameters.reflection_selection.method = "use_all"
    return parameters


@pytest.fixture
def mock_scaling_component():
    """Mock scaling component to allow creation of a scaling model."""
    component = MagicMock()
    component.n_params = 2
    component.inverse_scales = flex.double([0.9, 1.1])
    return component


def mock_exp(mock_scaling_component, idval=0):
    """Mock experiments object for initialising a scaler."""

    def side_effect_config_table(*args):
        """Side effect to mock configure reflection table
        call during initialisation."""
        return args[0]

    exp = MagicMock()
    exp.identifier = str(idval)
    exp.scaling_model.components = {"scale": mock_scaling_component}
    exp.scaling_model.consecutive_refinement_order = ["scale"]
    exp.scaling_model.is_scaled = False
    exp.scaling_model.error_model = BasicErrorModel()
    exp.scaling_model.configure_reflection_table.side_effect = side_effect_config_table
    exp_dict = {
        "__id__": "crystal",
        "real_space_a": [1.0, 0.0, 0.0],
        "real_space_b": [0.0, 1.0, 0.0],
        "real_space_c": [0.0, 0.0, 2.0],
        "space_group_hall_symbol": " C 2y",
    }
    exp.crystal = Crystal.from_dict(exp_dict)
    exp.scan.get_oscillation.return_value = (0, 1.0)
    exp.beam.get_sample_to_source_direction.return_value = (0.0, 0.0, -1.0)
    exp.goniometer.get_rotation_axis.return_value = (0.0, 0.0, 1.0)
    return exp


def mock_explist_3exp(mock_scaling_component):
    """A mock experimentlist, containing one mock exp instance three times."""
    exp = [mock_exp(mock_scaling_component)]
    exp.append(mock_exp(mock_scaling_component))
    exp.append(mock_exp(mock_scaling_component))
    return exp


@pytest.fixture
def mock_scaled_exp():
    """A mock experiments object with scaling_model.is_scaled = True"""
    exp = Mock()
    exp.scaling_model.is_scaled = True
    return exp


@pytest.fixture
def mock_unscaled_exp():
    """A mock experiments object with scaling_model.is_scaled = False"""
    exp = Mock()
    exp.scaling_model.is_scaled = False
    return exp


@pytest.fixture
def mock_experimentlist(mock_scaled_exp, mock_unscaled_exp):
    """A mock experimentlist of mock scaled/unscaled mock exp."""
    explist = [
        mock_scaled_exp,
        mock_scaled_exp,
        mock_unscaled_exp,
        mock_scaled_exp,
        mock_unscaled_exp,
    ]
    return explist


def test_SingleScalerFactory(generated_param, refl_to_filter, mock_scaling_component):
    """Test the single scaler factory."""
    test_refl, exp = test_refl_and_exp(mock_scaling_component)
    # Test that all required attributes get added with standard params.
    assert all(
        (i not in test_refl) for i in ["inverse_scale_factor", "intensity", "variance"]
    )
    # Test default, (no split into free set)
    ss = SingleScalerFactory.create(generated_param, exp, test_refl)
    assert isinstance(ss, SingleScaler)
    assert all(
        i in ss.reflection_table
        for i in ["inverse_scale_factor", "intensity", "variance"]
    )

    # Test reflection filtering
    rt = SingleScalerFactory.filter_bad_reflections(refl_to_filter)
    assert list(rt.get_flags(rt.flags.excluded_for_scaling)) == [
        True,
        True,
        False,
        False,
        False,
        False,
    ]


def test_selection_of_profile_or_summation_intensities(
    generated_param, prf_sum_refl_to_filter, mock_scaling_component
):
    _, exp = test_refl_and_exp(mock_scaling_component)
    # Test that all required attributes get added with standard params.
    assert all(
        (i not in prf_sum_refl_to_filter)
        for i in ["inverse_scale_factor", "intensity", "variance"]
    )
    # Test default, (no split into free set)
    ss = SingleScalerFactory.create(generated_param, exp, prf_sum_refl_to_filter)
    assert isinstance(ss, SingleScaler)
    rt = ss.reflection_table
    assert all(i in rt for i in ["inverse_scale_factor", "intensity", "variance"])
    assert list(rt.get_flags(rt.flags.excluded_for_scaling)) == [
        False,
        True,
        False,
        False,
        False,
    ]
    # test correct initial intensities have been chosen - should be prf then sum
    assert list(rt["intensity"]) == [11.0, 2.0, 3.0, 14.0, 15.0]


def test_TargetScalerFactory(generated_param, mock_scaling_component):
    """Test the target scaler factory."""

    refl_list, explist = test_refl_and_exp_list(mock_scaling_component, 3)

    # Test standard initialisation.
    assert generated_param.scaling_options.use_free_set is False  # just to check
    explist[0].scaling_model.is_scaled = True
    explist[1].scaling_model.is_scaled = True
    target = TargetScalerFactory.create(generated_param, explist, refl_list)
    assert isinstance(target, TargetScaler)
    assert len(target.single_scalers) == 2
    assert len(target.unscaled_scalers) == 1
    assert set(target.single_scalers[0].reflection_table["id"]) == {0}
    assert set(target.single_scalers[1].reflection_table["id"]) == {1}
    assert set(target.unscaled_scalers[0].reflection_table["id"]) == {2}

    # Now test converting targetscaler to multiscaler
    multiscaler = MultiScalerFactory.create_from_targetscaler(target)
    assert isinstance(multiscaler, MultiScaler)
    assert len(multiscaler.single_scalers) == 3

    # Test for correct initialisation when scaling against a target model.
    generated_param.scaling_options.reference = True
    target = TargetScalerFactory.create_for_target_against_reference(
        generated_param, explist, refl_list
    )
    assert isinstance(target.single_scalers[0], NullScaler)

    # This time make one dataset bad, and check it gets removed
    refl_list, explist = test_refl_and_exp_list(mock_scaling_component, 3)
    generated_param.scaling_options.reference = False
    refl_list[1].unset_flags(flex.bool(4, True), refl_list[1].flags.integrated_prf)
    refl_list[1].unset_flags(flex.bool(4, True), refl_list[1].flags.integrated_sum)
    explist[0].scaling_model.is_scaled = True
    target = TargetScalerFactory.create(generated_param, explist, refl_list)
    assert isinstance(target, TargetScaler)
    assert len(target.single_scalers) == 1
    assert len(target.unscaled_scalers) == 1
    assert set(target.single_scalers[0].reflection_table["id"]) == {0}
    assert set(target.unscaled_scalers[0].reflection_table["id"]) == {2}

    refl_list, explist = test_refl_and_exp_list(mock_scaling_component, 3)
    refl_list[0].unset_flags(flex.bool(4, True), refl_list[0].flags.integrated_prf)
    refl_list[0].unset_flags(flex.bool(4, True), refl_list[0].flags.integrated_sum)
    explist[0].scaling_model.is_scaled = True
    explist[1].scaling_model.is_scaled = True
    target = TargetScalerFactory.create(generated_param, explist, refl_list)
    assert isinstance(target, TargetScaler)
    assert len(target.single_scalers) == 1
    assert len(target.unscaled_scalers) == 1
    assert set(target.single_scalers[0].reflection_table["id"]) == {1}
    assert set(target.unscaled_scalers[0].reflection_table["id"]) == {2}


def test_MultiScalerFactory(generated_param, mock_scaling_component, refl_list):
    """Test the MultiScalerFactory."""

    refl_list, explist = test_refl_and_exp_list(mock_scaling_component, 3)

    multiscaler = MultiScalerFactory.create(generated_param, explist, refl_list)
    assert isinstance(multiscaler, MultiScaler)
    assert len(multiscaler.single_scalers) == 3
    for i in range(3):
        assert set(multiscaler.single_scalers[i].reflection_table["id"]) == {i}

    # This time make one dataset bad, and check it gets removed
    r1 = generated_refl(not_integrated=True)
    r2 = generated_refl()
    r3 = generated_refl()
    new_list = [r1, r2, r3]
    multiscaler = MultiScalerFactory.create(
        generated_param, mock_explist_3exp(mock_scaling_component), new_list
    )
    assert isinstance(multiscaler, MultiScaler)
    assert len(multiscaler.single_scalers) == 2
    r1 = multiscaler.single_scalers[0].reflection_table
    assert list(r1.get_flags(r1.flags.integrated)) == [True, True, False, False]
    r2 = multiscaler.single_scalers[1].reflection_table
    assert list(r2.get_flags(r2.flags.integrated)) == [True, True, False, False]


def test_scaler_factory_helper_functions(
    mock_experimentlist, generated_param, refl_list, mock_scaling_component
):
    """Test the helper functions."""

    test_refl, exp = test_refl_and_exp(mock_scaling_component)

    # Test create_scaler
    # Test case for single refl and exp
    scaler = create_scaler(generated_param, [exp], [test_refl])
    assert isinstance(scaler, SingleScaler)

    # If none or allscaled
    explist = mock_explist_3exp(mock_scaling_component)
    scaler = create_scaler(generated_param, explist, refl_list)
    assert isinstance(scaler, MultiScaler)

    explist[0].scaling_model.is_scaled = False
    # ^ changes all in list as same instance of exp.
    scaler = create_scaler(
        generated_param, mock_explist_3exp(mock_scaling_component), refl_list
    )
    assert isinstance(scaler, MultiScaler)

    # If only some scaled
    explist = []
    explist.append(mock_exp(mock_scaling_component))
    explist.append(mock_exp(mock_scaling_component))
    explist[1].scaling_model.is_scaled = True
    r1 = generated_refl()
    r2 = generated_refl()
    refl_list = [r1, r2]
    scaler = create_scaler(generated_param, explist, refl_list)
    assert isinstance(scaler, TargetScaler)

    # If no reflections passed in.
    with pytest.raises(ValueError):
        scaler = create_scaler(
            generated_param, mock_explist_3exp(mock_scaling_component), []
        )



from __future__ import annotations

import random

from dials.array_family import flex


class Simulator:
    """Class to help with simulation from reciprocal space."""

    def __init__(self, experiment, sigma_b, sigma_m, n_sigma):
        """Initialise with models and parameters."""
        self.experiment = experiment
        self.sigma_b = sigma_b
        self.sigma_m = sigma_m
        self.n_sigma = n_sigma

    def with_given_intensity(self, N, In, Ba, Bb, Bc, Bd):
        """Generate reflections with a given intensity and background."""
        return self.with_individual_given_intensity(
            N,
            flex.int(N, In),
            flex.int(N, Ba),
            flex.int(N, Bb),
            flex.int(N, Bc),
            flex.int(N, Bd),
        )

    def with_random_intensity(self, N, Imax, Bamax, Bbmax, Bcmax, Bdmax):
        """Generate reflections with a random intensity and background."""
        if Imax == 0:
            In = flex.size_t(N).as_int()
        else:
            In = flex.random_size_t(N, Imax).as_int()
        if Bamax == 0:
            Ba = flex.size_t(N).as_int()
        else:
            Ba = flex.random_size_t(N, Bamax).as_int()
        if Bbmax == 0:
            Bb = flex.size_t(N).as_int()
        else:
            Bb = flex.random_size_t(N, Bbmax).as_int()
        if Bcmax == 0:
            Bc = flex.size_t(N).as_int()
        else:
            Bc = flex.random_size_t(N, Bcmax).as_int()
        if Bdmax == 0:
            Bd = flex.size_t(N).as_int()
        else:
            Bd = flex.random_size_t(N, Bdmax).as_int()
        return self.with_individual_given_intensity(N, In, Ba, Bb, Bc, Bd)

    def with_individual_given_intensity(self, N, In, Ba, Bb, Bc, Bd):
        """Generate reflections with given intensity and background."""
        from dials.algorithms.simulation import simulate_reciprocal_space_gaussian
        from dials.algorithms.simulation.generate_test_reflections import (
            random_background_plane2,
        )
        from dials.util.command_line import ProgressBar

        # Check the lengths
        assert N == len(In)
        assert N == len(Ba)
        assert N == len(Bb)
        assert N == len(Bc)
        assert N == len(Bd)

        # Generate some predictions
        refl = self.generate_predictions(N)

        # Calculate the signal
        progress = ProgressBar(title=f"Calculating signal for {len(refl)} reflections")
        s1 = refl["s1"]
        phi = refl["xyzcal.mm"].parts()[2]
        bbox = refl["bbox"]
        shoebox = refl["shoebox"]
        m = int(len(refl) / 100)
        I_exp = flex.double(len(refl), 0)
        for i in range(len(refl)):
            if In[i] > 0:
                data = shoebox[i].data.as_double()
                I_exp[i] = simulate_reciprocal_space_gaussian(
                    self.experiment.beam,
                    self.experiment.detector,
                    self.experiment.goniometer,
                    self.experiment.scan,
                    self.sigma_b,
                    self.sigma_m,
                    s1[i],
                    phi[i],
                    bbox[i],
                    In[i],
                    data,
                    shoebox[i].mask,
                )
                shoebox[i].data = data.as_float()
            if i % m == 0:
                progress.update(100.0 * float(i) / len(refl))
        progress.finished(f"Calculated signal impacts for {len(refl)} reflections")

        # Calculate the background
        progress = ProgressBar(
            title=f"Calculating background for {len(refl)} reflections"
        )
        for l in range(len(refl)):
            background = flex.float(flex.grid(shoebox[l].size()), 0.0)
            random_background_plane2(background, Ba[l], Bb[l], Bc[l], Bd[l])
            shoebox[l].data += background
            shoebox[l].background = background
            if l % m == 0:
                progress.update(100.0 * float(l) / len(refl))
            progress.update(100.0 * float(l) / len(refl))
        progress.finished(f"Calculated background for {len(refl)} reflections")

        ## Calculate the expected intensity by monte-carlo integration
        # progress = ProgressBar(title='Integrating expected signal for %d reflections' % len(refl))
        # s1 = refl['s1']
        # phi = refl['xyzcal.mm'].parts()[2]
        # bbox = refl['bbox']
        # shoebox = refl['shoebox']
        # I_exp = flex.double(len(refl), 0)
        # m = int(len(refl) / 100)
        # for i in range(len(refl)):
        # if In[i] > 0:
        # I_exp[i] = integrate_reciprocal_space_gaussian(
        # self.experiment.beam,
        # self.experiment.detector,
        # self.experiment.goniometer,
        # self.experiment.scan,
        # self.sigma_b,
        # self.sigma_m,
        # s1[i],
        # phi[i],
        # bbox[i],
        # 10000,
        # shoebox[i].mask) / 10000.0
        # if i % m == 0:
        # progress.update(100.0 * float(i) / len(refl))
        # progress.finished('Integrated expected signal impacts for %d reflections' % len(refl))

        # Save the expected intensity and background
        refl["intensity.sim"] = In
        refl["background.sim.a"] = Ba
        refl["background.sim.b"] = Bb
        refl["background.sim.c"] = Bc
        refl["background.sim.d"] = Bd
        refl["intensity.exp"] = I_exp

        # Return the reflections
        return refl

    def generate_predictions(self, N):
        """Generate some reflections."""
        from dials.algorithms import filtering
        from dials.algorithms.profile_model.gaussian_rs import MaskCalculator3D
        from dials.algorithms.profile_model.gaussian_rs import Model as ProfileModel
        from dials.algorithms.shoebox import MaskCode
        from dials.util.command_line import Command

        # Set the profile model
        self.experiment.profile = ProfileModel(
            None, self.n_sigma, self.sigma_b, self.sigma_m
        )

        # Generate a list of reflections
        refl = flex.reflection_table.from_predictions(self.experiment)
        refl["id"] = flex.int(len(refl), 0)

        # Filter by zeta
        zeta = 0.05
        Command.start(f"Filtering by zeta >= {zeta:f}")
        mask = filtering.by_zeta(
            self.experiment.goniometer, self.experiment.beam, refl["s1"], zeta
        )
        refl.del_selected(~mask)
        Command.end(f"Filtered {len(refl)} reflections by zeta >= {zeta:f}")

        # Compute the bounding box
        refl.compute_bbox([self.experiment])
        index = []
        image_size = self.experiment.detector[0].get_image_size()
        array_range = self.experiment.scan.get_array_range()
        bbox = refl["bbox"]
        for i in range(len(refl)):
            x0, x1, y0, y1, z0, z1 = bbox[i]
            if (
                x0 < 0
                or x1 > image_size[0]
                or y0 < 0
                or y1 > image_size[1]
                or z0 < array_range[0]
                or z1 > array_range[1]
            ):
                index.append(i)
        refl.del_selected(flex.size_t(index))

        # Sample if specified
        index = random.sample(range(len(refl)), N)
        refl = refl.select(flex.size_t(index))

        # Compute the bounding box
        # Create a load of shoeboxes
        Command.start(f"Creating shoeboxes for {len(refl)} reflections")
        refl["shoebox"] = flex.shoebox(refl["panel"], refl["bbox"])
        refl["shoebox"].allocate_with_value(MaskCode.Valid)
        Command.end(f"Created shoeboxes for {len(refl)} reflections")

        # Get the function object to mask the foreground
        Command.start(f"Masking Foreground for {len(refl)} reflections")
        mask_foreground = MaskCalculator3D(
            self.experiment.beam,
            self.experiment.detector,
            self.experiment.goniometer,
            self.experiment.scan,
            self.n_sigma * self.sigma_b,
            self.n_sigma * self.sigma_m,
        )

        # Mask the foreground
        mask_foreground(
            refl["shoebox"], refl["s1"], refl["xyzcal.px"].parts()[2], refl["panel"]
        )
        Command.end(f"Masked foreground for {len(refl)} reflections")

        # Return the reflections
        return refl



from __future__ import annotations

import logging
import pickle

from dials.util.phil import parse

logger = logging.getLogger(__name__)


phil_scope = parse(
    """

  min_count = 5
    .type = int(value_min=1)

  nsigma = 6
    .type = float(value_min=0)

  sigma = 0.5
    .type = float(value_min=0)

  kernel_size = 9
    .type = int

  niter = 10
    .type = int
"""
)


class Modeller:
    def __init__(
        self, beam, detector, min_count=5, nsigma=6, sigma=0.5, kernel_size=9, niter=10
    ):
        from dials.algorithms.background.gmodel import PixelFilter

        self.beam = beam
        self.detector = detector

        self.min_count = min_count
        self.nsigma = nsigma
        self.sigma = sigma
        self.kernel_size = kernel_size
        self.niter = niter

        width, height = detector[0].get_image_size()

        self._filter = PixelFilter(height, width)

        self.detector_mask = None

    def add_image(self, frame, image, mask, reflections):

        height, width = image.all()

        _, _, _, _, z0, z1 = reflections["bbox"].parts()
        selection = (z0 <= frame) & (z1 > frame)
        subset = reflections.select(selection)
        sbox_mask = subset["shoebox"].apply_background_mask(frame, 1, (height, width))

        if self.detector_mask is None:
            self.detector_mask = mask

        mask = mask & sbox_mask

        self._filter.add(image, mask)

    def compute(self):
        from dials.algorithms.background.gmodel import FillGaps
        from dials.algorithms.image.fill_holes import simple_fill

        result = self._filter.compute(self.min_count, self.nsigma)

        data = result.data()
        mask = result.mask()

        data = simple_fill(data, mask)

        fill = FillGaps(self.beam, self.detector[0])

        mask = mask.as_1d().as_int()
        mask = mask - (~self.detector_mask).as_1d().as_int()
        mask.reshape(data.accessor())

        fill(data, mask, self.sigma, self.kernel_size, self.niter)

        return data


class Creator:
    def __init__(self, experiment, params):

        self.modeller = None
        self.background = None
        self.experiment = experiment
        self.params = params

    def initialized(self):
        return self.modeller is not None

    def finalized(self):
        return self.background is not None

    def initialize(self):
        assert not self.initialized()
        self.modeller = Modeller(
            self.experiment.beam,
            self.experiment.detector,
            self.params.min_count,
            self.params.nsigma,
        )

    def finalize(self):
        assert self.initialized()
        self.background = self.modeller.compute()
        self.modeller = None
        if self.params.debug.output:
            filename = self.params.debug.filename
            logger.info("Writing background model to %s", filename)
            with open(filename, "wb") as outfile:
                pickle.dump(self.background, outfile, protocol=pickle.HIGHEST_PROTOCOL)

    def compute(self, reflections):
        from dials.algorithms.background.gmodel import Fitter
        from dials.array_family import flex

        assert self.finalized()
        fitter = Fitter(self.background)
        scale = fitter(reflections["shoebox"])
        success = scale >= 0
        mean = flex.double(
            [flex.mean(sbox.background) for sbox in reflections["shoebox"]]
        )
        reflections["background.mean"] = mean
        reflections["background.scale"] = scale
        reflections.set_flags(~success, reflections.flags.dont_integrate)
        return success



"""Test the plotting module from systematic_absences."""

from __future__ import annotations

from dials.algorithms.symmetry.absences.plots import color_axis_data, plot_screw_axes


def test_color_axis_data():
    """Test the helper function used to color the data."""
    colors = color_axis_data("41c", list(range(1, 9)))
    assert colors == [0, 0, 0, 1, 0, 0, 0, 1]
    colors = color_axis_data("42c", list(range(1, 9)))
    assert colors == [0, 1, 0, 1, 0, 1, 0, 1]
    colors = color_axis_data("31c", list(range(1, 7)))
    assert colors == [0, 0, 1, 0, 0, 1]
    colors = color_axis_data("21c", list(range(1, 7)))
    assert colors == [0, 1, 0, 1, 0, 1]
    colors = color_axis_data("61c", list(range(1, 7)))
    assert colors == [0, 0, 0, 0, 0, 1]
    colors = color_axis_data("62c", list(range(1, 7)))
    assert colors == [0, 0, 1, 0, 0, 1]
    colors = color_axis_data("63c", list(range(1, 7)))
    assert colors == [0, 1, 0, 1, 0, 1]


def test_plot_screw_axes():
    """Test that the plotting function returns plots with data."""
    data = {
        "miller_axis_vals": [1, 2, 3, 4, 5, 6],
        "i_over_sigma": [0.5, 20.0, 0.2, 10.0, 0.4, 15.0],
        "intensities": [0.5, 20.0, 0.2, 10.0, 0.4, 15.0],
        "sigmas": [1.0] * 6,
        "fourier_space_data": {},
    }
    plots = plot_screw_axes({"21a": data})
    for pl in plots.values():
        assert pl["data"][0]["x"]
        assert pl["data"][0]["y"]



from __future__ import annotations

from dials_util_ext import *  # noqa: F403; lgtm

__all__ = (  # noqa: F405
    "ResolutionMaskGenerator",
    "add_dials_batches",
    "dials_u_to_mosflm",
    "ostream",
    "scale_down_array",
    "streambuf",
)



from __future__ import annotations

from collections import namedtuple

from scitbx import matrix, sparse

from dials.algorithms.refinement import DialsRefineConfigError
from dials.array_family import flex

"""The PredictionParameterisation class ties together parameterisations for
individual experimental models: beam, crystal orientation, crystal unit cell
detector and goniometer into a parameterisation for the reflection prediction
equation. Methods are provided to get and set the global list of parameters in
a defined order and calculate the gradients of the predicted centroids with
respect to the global parameters. A concrete version of this class is provided
for scan-static reflection prediction for rotation data, where the predicted
centroid is expressed as X, Y, phi. Other versions of the class are defined
elsewhere."""

ParamSet = namedtuple(
    "ParamSet",
    ["beam_param", "xl_ori_param", "xl_uc_param", "det_param", "gonio_param"],
)


class PredictionParameterisation:
    """
    Abstract interface for a class that groups together model parameterisations
    relating to diffraction geometry and provides:

    * A list of all free parameters concatenated from each of the models, with a
      getter and setter method that delegates to the contained models
    * Derivatives of the reflection prediction equation with respect to each of
      these free parameters

    Derived classes determine whether the reflection prediction equation is
    expressed in detector space (X, Y, phi) or orthogonalised reciprocal space.

    It is assumed that the provided model parameterisations will be one of five
    types:

    * Detector parameterisation
    * Beam parameterisation
    * Crystal orientation parameterisation
    * Crystal unit cell parameterisation
    * Goniometer setting matrix parameterisation

    No more than one of each type must be supplied for each Experiment. Omitting
    a parameterisation for any of the types is allowed, which implies that the
    attached model is fixed in refinement of that Experiment.

    A class implementing PredictionParameterisation is used by a Refinery
    object directly, which takes the list of parameters, and indirectly via a
    Target function object, which takes the list of derivatives and composes the
    derivatives of a Target function from them.
    """

    def __init__(
        self,
        experiments,
        detector_parameterisations=None,
        beam_parameterisations=None,
        xl_orientation_parameterisations=None,
        xl_unit_cell_parameterisations=None,
        goniometer_parameterisations=None,
    ):

        if detector_parameterisations is None:
            detector_parameterisations = []
        if beam_parameterisations is None:
            beam_parameterisations = []
        if xl_orientation_parameterisations is None:
            xl_orientation_parameterisations = []
        if xl_unit_cell_parameterisations is None:
            xl_unit_cell_parameterisations = []
        if goniometer_parameterisations is None:
            goniometer_parameterisations = []

        # References to the underlying models
        self._experiments = experiments

        # Keep references to all parameterised models
        self._detector_parameterisations = detector_parameterisations
        self._beam_parameterisations = beam_parameterisations
        self._xl_orientation_parameterisations = xl_orientation_parameterisations
        self._xl_unit_cell_parameterisations = xl_unit_cell_parameterisations
        self._goniometer_parameterisations = goniometer_parameterisations

        self._update()

    def _update(self):
        """Update information derived from the parameterisations"""

        # Check there are free parameters to refine
        self._length = self._len()
        if self._length == 0:
            raise DialsRefineConfigError("There are no free parameters for refinement")

        # Calculate Experiment to parameterisation mapping
        e2bp = {
            ids: i
            for i, p in enumerate(self._beam_parameterisations)
            for ids in p.get_experiment_ids()
        }
        e2xop = {
            ids: i
            for i, p in enumerate(self._xl_orientation_parameterisations)
            for ids in p.get_experiment_ids()
        }
        e2xucp = {
            ids: i
            for i, p in enumerate(self._xl_unit_cell_parameterisations)
            for ids in p.get_experiment_ids()
        }
        e2dp = {
            ids: i
            for i, p in enumerate(self._detector_parameterisations)
            for ids in p.get_experiment_ids()
        }
        e2gp = {
            ids: i
            for i, p in enumerate(self._goniometer_parameterisations)
            for ids in p.get_experiment_ids()
        }

        self._exp_to_param = {
            i: ParamSet(
                e2bp.get(i), e2xop.get(i), e2xucp.get(i), e2dp.get(i), e2gp.get(i)
            )
            for i, _ in enumerate(self._experiments)
        }

    # accessors for the lists of parameterisations of different types
    def get_detector_parameterisations(self):
        return self._detector_parameterisations

    def get_beam_parameterisations(self):
        return self._beam_parameterisations

    def get_crystal_orientation_parameterisations(self):
        return self._xl_orientation_parameterisations

    def get_crystal_unit_cell_parameterisations(self):
        return self._xl_unit_cell_parameterisations

    def get_goniometer_parameterisations(self):
        return self._goniometer_parameterisations

    def _len(self):

        length = 0
        for model in self._detector_parameterisations:
            length += model.num_free()
        for model in self._beam_parameterisations:
            length += model.num_free()
        for model in self._xl_orientation_parameterisations:
            length += model.num_free()
        for model in self._xl_unit_cell_parameterisations:
            length += model.num_free()
        for model in self._goniometer_parameterisations:
            length += model.num_free()

        return length

    def __len__(self):
        return self._length

    def get_param_vals(self):
        """Return a concatenated list of parameters from each of the components
        in the global model"""

        global_p_list = []
        if self._detector_parameterisations:
            det_plists = [x.get_param_vals() for x in self._detector_parameterisations]
            params = [x for l in det_plists for x in l]
            global_p_list.extend(params)

        if self._beam_parameterisations:
            src_plists = [x.get_param_vals() for x in self._beam_parameterisations]
            params = [x for l in src_plists for x in l]
            global_p_list.extend(params)

        if self._xl_orientation_parameterisations:
            xlo_plists = [
                x.get_param_vals() for x in self._xl_orientation_parameterisations
            ]
            params = [x for l in xlo_plists for x in l]
            global_p_list.extend(params)

        if self._xl_unit_cell_parameterisations:
            xluc_plists = [
                x.get_param_vals() for x in self._xl_unit_cell_parameterisations
            ]
            params = [x for l in xluc_plists for x in l]
            global_p_list.extend(params)

        if self._goniometer_parameterisations:
            gon_plists = [
                x.get_param_vals() for x in self._goniometer_parameterisations
            ]
            params = [x for l in gon_plists for x in l]
            global_p_list.extend(params)

        return global_p_list

    def get_param_names(self):
        """Return a list of the names of parameters in the order they are
        concatenated. Use 1-based indexing for indices in the names."""

        param_names = []
        for p in self._detector_parameterisations:
            prefix = p.model_identifier
            param_names.extend([prefix + x for x in p.get_param_names()])

        for p in self._beam_parameterisations:
            prefix = p.model_identifier
            param_names.extend([prefix + x for x in p.get_param_names()])

        for p in self._xl_orientation_parameterisations:
            prefix = p.model_identifier
            param_names.extend([prefix + x for x in p.get_param_names()])

        for p in self._xl_unit_cell_parameterisations:
            prefix = p.model_identifier
            param_names.extend([prefix + x for x in p.get_param_names()])

        for p in self._goniometer_parameterisations:
            prefix = p.model_identifier
            param_names.extend([prefix + x for x in p.get_param_names()])

        return param_names

    def _modify_parameters(self, vals, set_vals=False, set_esds=False, set_fix=False):
        assert [set_vals, set_esds, set_fix].count(True) == 1
        assert len(vals) == len(self)
        it = iter(vals)

        for model in (
            self._detector_parameterisations
            + self._beam_parameterisations
            + self._xl_orientation_parameterisations
            + self._xl_unit_cell_parameterisations
            + self._goniometer_parameterisations
        ):
            tmp = [next(it) for i in range(model.num_free())]
            if set_esds:
                model.set_param_esds(tmp)
            elif set_vals:
                model.set_param_vals(tmp)
            elif set_fix:
                current_fixes = model.get_fixed()
                free_indices = [i for i, e in enumerate(current_fixes) if not e]
                assert len(free_indices) == model.num_free()
                for i, fix in zip(free_indices, tmp):
                    if fix:
                        current_fixes[i] = True
                model.set_fixed(current_fixes)

    def set_param_vals(self, vals):
        """Set the parameter values of the contained models to the values in
        vals. This list must be of the same length as the result of get_param_vals
        and must contain the parameter values in the same order."""

        self._modify_parameters(vals, set_vals=True)

    def set_param_esds(self, esds):
        """Set the estimated standard deviations of parameter values of the
        contained models to the values in esds. This list must be of the same length
        as the result of get_param_vals and must contain the parameter values in the
        same order."""

        self._modify_parameters(esds, set_esds=True)

    def fix_params(self, fix):
        """Fix the parameters according to the boolean values in fix. This list
        must be of the same length as the result of get_param_vals and will
        fix parameters at the positions that are True within the list"""

        self._modify_parameters(fix, set_fix=True)
        self._update()

    def calculate_model_state_uncertainties(self, var_cov):
        """Take a variance-covariance matrix of all free parameters (probably
        calculated by a minimisation engine). For each parameterisation in the
        global model, extract the subset of this matrix for the associated block
        of parameters. Pass this on to the relevant model parameterisation to
        calculate its own uncertainty of state."""

        i = 0
        for model in (
            self._detector_parameterisations
            + self._beam_parameterisations
            + self._xl_orientation_parameterisations
            + self._xl_unit_cell_parameterisations
            + self._goniometer_parameterisations
        ):
            n = model.num_free()
            sub = var_cov.matrix_copy_block(i, i, n, n)
            state_covs = model.calculate_state_uncertainties(sub)
            if state_covs is None:
                continue
            if len(state_covs) == 1:
                model.set_state_uncertainties(state_covs[0])
            else:
                for i_state, state_cov in enumerate(state_covs):
                    model.set_state_uncertainties(state_cov, multi_state_elt=i_state)
            i += n

    def get_gradients(self, reflections, callback=None):
        """Calculate gradients of the prediction formula with respect to each
        of the parameters of the contained models, for all of the reflections.
        This method sets up required quantities relevant to the current step of
        refinement and then loops over the parameterisations of each type extending
        a results list each time."""

        # Set up arrays of quantities of interest for each reflection
        self._nref = len(reflections)
        self._D = flex.mat3_double(self._nref)
        self._s0 = flex.vec3_double(self._nref)
        self._U = flex.mat3_double(self._nref)
        self._B = flex.mat3_double(self._nref)
        self._axis = flex.vec3_double(self._nref)
        self._fixed_rotation = flex.mat3_double(self._nref)
        self._setting_rotation = flex.mat3_double(self._nref)

        # Set up experiment to index mapping
        self._experiment_to_idx = []

        # Populate values in these arrays
        for iexp, exp in enumerate(self._experiments):

            sel = reflections["id"] == iexp
            isel = sel.iselection()
            self._experiment_to_idx.append(isel)
            subref = reflections.select(sel)
            states = self._get_model_data_for_experiment(exp, subref)

            self._D.set_selected(sel, states["D"])
            self._s0.set_selected(sel, states["s0"])
            self._U.set_selected(sel, states["U"])
            self._B.set_selected(sel, states["B"])
            if exp.goniometer:
                self._setting_rotation.set_selected(sel, states["S"])
                self._axis.set_selected(sel, exp.goniometer.get_rotation_axis_datum())
                self._fixed_rotation.set_selected(
                    sel, exp.goniometer.get_fixed_rotation()
                )

        # Other derived values
        self._h = reflections["miller_index"].as_vec3_double()
        self._UB = self._U * self._B
        self._s1 = reflections["s1"]
        self._pv = self._D * self._s1  # 'projection vector' for the ray along s1.

        # Quantities derived from pv, precalculated for efficiency
        u, v, w = self._pv.parts()
        assert w.all_ne(0)
        self._w_inv = 1.0 / w
        self._u_w_inv = u * self._w_inv
        self._v_w_inv = v * self._w_inv

        # Reset a pointer to the parameter number
        self._iparam = 0

        # Do additional setup specified by derived classes
        self._local_setup(reflections)

        # Set up empty list in which to store gradients
        results = []

        # loop over detector parameterisations and extend results
        results = self._grads_detector_loop(reflections, results, callback=callback)

        # loop over the beam parameterisations and extend results
        results = self._grads_beam_loop(reflections, results, callback=callback)

        # loop over the crystal orientation parameterisations and extend results
        results = self._grads_xl_orientation_loop(
            reflections, results, callback=callback
        )

        # loop over the crystal unit cell parameterisations and extend results
        results = self._grads_xl_unit_cell_loop(reflections, results, callback=callback)

        # loop over the goniometer parameterisations and extend results
        results = self._grads_goniometer_loop(reflections, results, callback=callback)

        return results

    @staticmethod
    def _extend_gradient_vectors(results, m, n, keys=("dX_dp", "dY_dp", "dZ_dp")):
        """Extend results list by n empty results. These will each be a dictionary
        indexed by the given keys. The value for each key will be an empty vector of
        size m, to store the derivatives of n parameters, for m reflections. This
        method may be overridden by a derived class to e.g. use sparse vectors"""

        new_results = []
        for i in range(n):
            result = {}
            for key in keys:
                result[key] = flex.double(m, 0.0)
            new_results.append(result)
        results.extend(new_results)

        return results

    def _get_model_data_for_experiment(self, experiment, reflections):
        """Helper function to return model data s0, U, B, D and S for a particular
        experiment. D is always returned as an array the same length as the
        reflections for the experiment, whereas here s0, U, B and S are returned as
        single matrices or vectors. In the scan-varying overload these will all be
        arrays."""

        # D matrix array
        D = flex.mat3_double(len(reflections))
        panels = reflections["panel"]
        for ipanel, D_mat in enumerate([p.get_D_matrix() for p in experiment.detector]):
            sel = panels == ipanel
            D.set_selected(sel, D_mat)

        result = {
            "s0": experiment.beam.get_s0(),
            "U": matrix.sqr(experiment.crystal.get_U()),
            "B": matrix.sqr(experiment.crystal.get_B()),
            "D": D,
        }

        # If a goniometer is present, get the setting matrix too
        if experiment.goniometer:
            result["S"] = matrix.sqr(experiment.goniometer.get_setting_rotation())

        return result

    # The detector derivatives calculation is shared by scans and stills type
    # prediction, so this method is here, in the base class.
    def _detector_derivatives(
        self, isel, panel_id, parameterisation=None, dd_ddet_p=None, reflections=None
    ):
        """Helper function to convert derivatives of the detector state to
        derivatives of the vector pv. Derivatives that would all be null vectors
        are replaced with None"""

        # Get required data
        pv = self._pv.select(isel)
        D = self._D.select(isel)

        if dd_ddet_p is None:

            # get the derivatives of detector d matrix for this panel
            dd_ddet_p = parameterisation.get_ds_dp(
                multi_state_elt=panel_id, use_none_as_null=True
            )

            # replace explicit null derivatives with None
            dd_ddet_p = [
                None if e is None else flex.mat3_double(len(D), e.elems)
                for e in dd_ddet_p
            ]

        # calculate the derivative of pv for this parameter
        dpv_ddet_p = [
            der if der is None else (D * (der * -1.0)) * pv for der in dd_ddet_p
        ]

        return dpv_ddet_p

    # The other model derivatives are different for scans and stills. Empty
    # implementations are declared here
    def _beam_derivatives(
        self, isel, parameterisation=None, ds0_dbeam_p=None, reflections=None
    ):
        pass

    def _xl_orientation_derivatives(
        self, isel, parameterisation=None, dU_dxlo_p=None, reflections=None
    ):
        pass

    def _xl_unit_cell_derivatives(
        self, isel, parameterisation=None, dB_dxluc_p=None, reflections=None
    ):
        pass

    def _goniometer_derivatives(
        self, isel, parameterisation=None, dS_dgon_p=None, reflections=None
    ):
        pass

    def _grads_detector_loop(self, reflections, results, callback=None):
        """Loop over all detector parameterisations, calculate gradients and extend
        the results"""

        # loop over the detector parameterisations
        for dp in self._detector_parameterisations:

            # Determine (sub)set of reflections affected by this parameterisation
            isel = flex.size_t()
            for exp_id in dp.get_experiment_ids():
                isel.extend(self._experiment_to_idx[exp_id])

            # Access the detector model being parameterised
            detector = dp.get_model()

            # Get panel numbers of the affected reflections
            panel = reflections["panel"].select(isel)

            # Extend derivative vectors for this detector parameterisation
            results = self._extend_gradient_vectors(
                results, self._nref, dp.num_free(), keys=self._grad_names
            )

            # loop through the panels in this detector
            for panel_id, _ in enumerate(detector):

                # get the right subset of array indices to set for this panel
                sub_isel = isel.select(panel == panel_id)
                if len(sub_isel) == 0:
                    # if no reflections intersect this panel, skip calculation
                    continue

                dpv_ddet_p = self._detector_derivatives(
                    sub_isel, panel_id, parameterisation=dp, reflections=reflections
                )

                # convert to dX/dp, dY/dp and assign the elements of the vectors
                # corresponding to this experiment and panel
                sub_w_inv = self._w_inv.select(sub_isel)
                sub_u_w_inv = self._u_w_inv.select(sub_isel)
                sub_v_w_inv = self._v_w_inv.select(sub_isel)
                dX_ddet_p, dY_ddet_p = self._calc_dX_dp_and_dY_dp_from_dpv_dp(
                    sub_w_inv, sub_u_w_inv, sub_v_w_inv, dpv_ddet_p
                )

                # use a local parameter index pointer because we set all derivatives
                # for this panel before moving on to the next
                iparam = self._iparam
                for dX, dY in zip(dX_ddet_p, dY_ddet_p):
                    if dX is not None:
                        try:
                            dX, indices = dX.data_and_indices
                            indices = sub_isel.select(indices)
                        except AttributeError:
                            indices = sub_isel
                        results[iparam][self._grad_names[0]].set_selected(indices, dX)
                    if dY is not None:
                        try:
                            dY, indices = dY.data_and_indices
                            indices = sub_isel.select(indices)
                        except AttributeError:
                            indices = sub_isel
                        results[iparam][self._grad_names[1]].set_selected(indices, dY)
                    # increment the local parameter index pointer
                    iparam += 1

            if callback is not None:
                iparam = self._iparam
                for i in range(dp.num_free()):
                    results[iparam] = callback(results[iparam])
                    iparam += 1

            # increment the parameter index pointer to the last detector parameter
            self._iparam += dp.num_free()

        return results

    def _grads_model_loop(
        self,
        parameterisations,
        reflections,
        results,
        callback=None,
        derivatives_fn=None,
    ):
        # loop over the parameterisations
        for p in parameterisations:

            # Determine (sub)set of reflections affected by this parameterisation
            isel = flex.size_t()
            for exp_id in p.get_experiment_ids():
                isel.extend(self._experiment_to_idx[exp_id])

            # Extend derivative vectors for this parameterisation
            results = self._extend_gradient_vectors(
                results, self._nref, p.num_free(), keys=self._grad_names
            )

            if len(isel) == 0:
                # if no reflections are in this experiment, skip calculation of
                # gradients, but must still process null gradients by a callback
                if callback:
                    for _ in range(p.num_free()):
                        results[self._iparam] = callback(results[self._iparam])
                        self._iparam += 1
                else:
                    self._iparam += p.num_free()
                continue

            w_inv = self._w_inv.select(isel)
            u_w_inv = self._u_w_inv.select(isel)
            v_w_inv = self._v_w_inv.select(isel)

            dpv_dp, dAngle_dp = derivatives_fn(
                isel, parameterisation=p, reflections=reflections
            )

            # convert to dX/dp, dY/dp and assign the elements of the vectors
            # corresponding to this experiment
            dX_dp, dY_dp = self._calc_dX_dp_and_dY_dp_from_dpv_dp(
                w_inv, u_w_inv, v_w_inv, dpv_dp
            )

            # Loop over gradients and paste them into the correct rows in the
            # results arrays, taking care of whether the gradients are flex
            # or SparseFlex arrays.
            for dX, dY, dAngle in zip(dX_dp, dY_dp, dAngle_dp):
                if dX is not None:
                    try:
                        dX, indices = dX.data_and_indices
                        indices = isel.select(indices)
                    except AttributeError:
                        indices = isel
                    results[self._iparam][self._grad_names[0]].set_selected(indices, dX)
                if dY is not None:
                    try:
                        dY, indices = dY.data_and_indices
                        indices = isel.select(indices)
                    except AttributeError:
                        indices = isel
                    results[self._iparam][self._grad_names[1]].set_selected(indices, dY)
                if dAngle is not None:
                    try:
                        dAngle, indices = dAngle.data_and_indices
                        indices = isel.select(indices)
                    except AttributeError:
                        indices = isel
                    results[self._iparam][self._grad_names[2]].set_selected(
                        indices, dAngle
                    )
                if callback is not None:
                    results[self._iparam] = callback(results[self._iparam])
                # increment the parameter index pointer
                self._iparam += 1

        return results

    def _grads_beam_loop(self, reflections, results, callback=None):
        """Loop over all beam parameterisations, calculate gradients and extend
        the results"""

        return self._grads_model_loop(
            self._beam_parameterisations,
            reflections,
            results,
            derivatives_fn=self._beam_derivatives,
            callback=callback,
        )

    def _grads_xl_orientation_loop(self, reflections, results, callback=None):
        """Loop over all crystal orientation parameterisations, calculate gradients
        and extend the results"""

        return self._grads_model_loop(
            self._xl_orientation_parameterisations,
            reflections,
            results,
            derivatives_fn=self._xl_orientation_derivatives,
            callback=callback,
        )

    def _grads_xl_unit_cell_loop(self, reflections, results, callback=None):
        """Loop over all crystal unit cell parameterisations, calculate gradients
        and extend the results"""

        return self._grads_model_loop(
            self._xl_unit_cell_parameterisations,
            reflections,
            results,
            derivatives_fn=self._xl_unit_cell_derivatives,
            callback=callback,
        )

    def _grads_goniometer_loop(self, reflections, results, callback=None):
        """Loop over all goniometer parameterisations, calculate gradients
        and extend the results"""

        return self._grads_model_loop(
            self._goniometer_parameterisations,
            reflections,
            results,
            derivatives_fn=self._goniometer_derivatives,
            callback=callback,
        )


class SparseGradientVectorMixin:
    """Mixin class to use sparse vectors for storage of gradients of the
    prediction formula"""

    @staticmethod
    def _extend_gradient_vectors(results, m, n, keys=("dX_dp", "dY_dp", "dZ_dp")):
        """Extend results list by n empty results. These will each be a dictionary
        indexed by the given keys. The value for each key will be an empty vector of
        size m, to store the derivatives of n parameters, for m reflections.
        This is the sparse vector version."""

        new_results = [{key: sparse.matrix_column(m) for key in keys} for _ in range(n)]
        results.extend(new_results)

        return results


class XYPhiPredictionParameterisation(PredictionParameterisation):

    _grad_names = ("dX_dp", "dY_dp", "dphi_dp")

    def _local_setup(self, reflections):
        """Setup additional attributes used in gradients calculation. These are
        specific to scans-type prediction parameterisations"""

        # Spindle rotation matrices for every reflection
        # R = self._axis.axis_and_angle_as_r3_rotation_matrix(phi)
        # R = flex.mat3_double(len(reflections))
        # NB for now use flex.vec3_double.rotate_around_origin each time I need the
        # rotation matrix R.

        # r is the reciprocal lattice vector, in the lab frame
        self._phi_calc = reflections["xyzcal.mm"].parts()[2]
        q = self._fixed_rotation * (self._UB * self._h)
        self._r = self._setting_rotation * q.rotate_around_origin(
            self._axis, self._phi_calc
        )

        # All of the derivatives of phi have a common denominator, given by
        # (e X r).s0, where e is the rotation axis. Calculate this once, here.
        self._e_X_r = (self._setting_rotation * self._axis).cross(self._r)
        self._e_r_s0 = (self._e_X_r).dot(self._s0)

        # Note that e_r_s0 -> 0 when the rotation axis, beam vector and
        # relp are coplanar. This occurs when a reflection just touches
        # the Ewald sphere.
        #
        # There is a relationship between e_r_s0 and zeta_factor.
        # Uncommenting the code below shows that
        # s0.(e X r) = zeta * |s X s0|

        # from dials.algorithms.profile_model.gaussian_rs import zeta_factor
        # from libtbx.test_utils import approx_equal
        # s = matrix.col(reflections['s1'][0])
        # z = zeta_factor(axis[0], s0[0], s)
        # ss0 = (s.cross(matrix.col(s0[0]))).length()
        # assert approx_equal(e_r_s0[0], z * ss0)

        # catch small values of e_r_s0
        e_r_s0_mag = flex.abs(self._e_r_s0)
        try:
            assert flex.min(e_r_s0_mag) > 1.0e-6
        except AssertionError as e:
            imin = flex.min_index(e_r_s0_mag)
            print("(e X r).s0 too small:")
            print("for", (e_r_s0_mag <= 1.0e-6).count(True), "reflections")
            print("out of", len(e_r_s0_mag), "total")
            print("such as", reflections["miller_index"][imin])
            print("with scattering vector", reflections["s1"][imin])
            print("where r =", self._r[imin])
            print("e =", self._axis[imin])
            print("s0 =", self._s0[imin])
            print("this reflection forms angle with the equatorial plane " "normal:")
            vecn = (
                matrix.col(self._s0[imin])
                .cross(matrix.col(self._axis[imin]))
                .normalize()
            )
            print(matrix.col(reflections["s1"][imin]).accute_angle(vecn))
            raise e

    def _beam_derivatives(
        self, isel, parameterisation=None, ds0_dbeam_p=None, reflections=None
    ):
        """helper function to extend the derivatives lists by derivatives of the
        beam parameterisations."""

        # Get required data
        r = self._r.select(isel)
        e_X_r = self._e_X_r.select(isel)
        e_r_s0 = self._e_r_s0.select(isel)
        D = self._D.select(isel)

        if ds0_dbeam_p is None:

            # get the derivatives of the beam vector wrt the parameters
            ds0_dbeam_p = parameterisation.get_ds_dp(use_none_as_null=True)

            ds0_dbeam_p = [
                None if e is None else flex.vec3_double(len(r), e.elems)
                for e in ds0_dbeam_p
            ]

        dphi_dp = []
        dpv_dp = []

        # loop through the parameters
        for der in ds0_dbeam_p:

            if der is None:
                dphi_dp.append(None)
                dpv_dp.append(None)
                continue

            # calculate the derivative of phi for this parameter
            dphi = (der.dot(r) / e_r_s0) * -1.0
            dphi_dp.append(dphi)

            # calculate the derivative of pv for this parameter
            dpv_dp.append(D * (e_X_r * dphi + der))

        return dpv_dp, dphi_dp

    def _xl_derivatives(self, isel, derivatives, b_matrix, parameterisation=None):
        """helper function to extend the derivatives lists by derivatives of
        generic parameterisations."""

        # Get required data
        axis = self._axis.select(isel)
        fixed_rotation = self._fixed_rotation.select(isel)
        setting_rotation = self._setting_rotation.select(isel)
        phi_calc = self._phi_calc.select(isel)
        h = self._h.select(isel)
        s1 = self._s1.select(isel)
        e_X_r = self._e_X_r.select(isel)
        e_r_s0 = self._e_r_s0.select(isel)
        if b_matrix:
            B = self._B.select(isel)
        else:
            U = self._U.select(isel)
        D = self._D.select(isel)

        if derivatives is None:
            # get derivatives of the B/U matrix wrt the parameters
            derivatives = [
                None if der is None else flex.mat3_double(len(isel), der.elems)
                for der in parameterisation.get_ds_dp(use_none_as_null=True)
            ]

        dphi_dp = []
        dpv_dp = []

        # loop through the parameters
        for der in derivatives:
            if der is None:
                dphi_dp.append(None)
                dpv_dp.append(None)
                continue

            # calculate the derivative of r for this parameter
            if b_matrix:
                tmp = fixed_rotation * (der * B * h)
            else:
                tmp = fixed_rotation * (U * der * h)
            dr = setting_rotation * tmp.rotate_around_origin(axis, phi_calc)

            # calculate the derivative of phi for this parameter
            dphi = -1.0 * dr.dot(s1) / e_r_s0
            dphi_dp.append(dphi)

            # calculate the derivative of pv for this parameter
            dpv_dp.append(D * (dr + e_X_r * dphi))

        return dpv_dp, dphi_dp

    def _xl_orientation_derivatives(
        self, isel, parameterisation=None, dU_dxlo_p=None, reflections=None
    ):
        """helper function to extend the derivatives lists by derivatives of the
        crystal orientation parameterisations"""
        return self._xl_derivatives(
            isel, dU_dxlo_p, b_matrix=True, parameterisation=parameterisation
        )

    def _xl_unit_cell_derivatives(
        self, isel, parameterisation=None, dB_dxluc_p=None, reflections=None
    ):
        """helper function to extend the derivatives lists by
        derivatives of the crystal unit cell parameterisations"""
        return self._xl_derivatives(
            isel, dB_dxluc_p, b_matrix=False, parameterisation=parameterisation
        )

    def _goniometer_derivatives(
        self, isel, parameterisation=None, dS_dgon_p=None, reflections=None
    ):
        """helper function to extend the derivatives lists by
        derivatives of the goniometer parameterisations"""

        # Get required data
        axis = self._axis.select(isel)
        fixed_rotation = self._fixed_rotation.select(isel)
        phi_calc = self._phi_calc.select(isel)
        h = self._h.select(isel)
        s1 = self._s1.select(isel)
        e_X_r = self._e_X_r.select(isel)
        e_r_s0 = self._e_r_s0.select(isel)
        UB = self._UB.select(isel)
        D = self._D.select(isel)

        if dS_dgon_p is None:

            # get derivatives of the setting matrix S wrt the parameters
            dS_dgon_p = [
                None if der is None else flex.mat3_double(len(isel), der.elems)
                for der in parameterisation.get_ds_dp(use_none_as_null=True)
            ]

        dphi_dp = []
        dpv_dp = []

        # loop through the parameters
        for der in dS_dgon_p:

            if der is None:
                dphi_dp.append(None)
                dpv_dp.append(None)
                continue

            # calculate the derivative of r for this parameter
            tmp = fixed_rotation * (UB * h)
            dr = der * tmp.rotate_around_origin(axis, phi_calc)

            # calculate the derivative of phi for this parameter
            dphi = -1.0 * dr.dot(s1) / e_r_s0
            dphi_dp.append(dphi)

            # calculate the derivative of pv for this parameter
            dpv_dp.append(D * (dr + e_X_r * dphi))

        return dpv_dp, dphi_dp

    @staticmethod
    def _calc_dX_dp_and_dY_dp_from_dpv_dp(w_inv, u_w_inv, v_w_inv, dpv_dp):
        """helper function to calculate positional derivatives from
        dpv_dp using the quotient rule"""

        dX_dp = []
        dY_dp = []

        for der in dpv_dp:
            if der is None:
                dX_dp.append(None)
                dY_dp.append(None)
            else:
                du_dp, dv_dp, dw_dp = der.parts()

                dX_dp.append(w_inv * (du_dp - dw_dp * u_w_inv))
                dY_dp.append(w_inv * (dv_dp - dw_dp * v_w_inv))

        return dX_dp, dY_dp


class XYPhiPredictionParameterisationSparse(
    SparseGradientVectorMixin, XYPhiPredictionParameterisation
):
    """A version of XYPhiPredictionParameterisation that uses a sparse matrix
    data structure for memory efficiency when there are a large number of
    Experiments"""

    # Explicitly delegate to the method from SparseGradientVectorMixin
    # (https://lgtm.com/rules/7860084/)
    @staticmethod
    def _extend_gradient_vectors(results, m, n, keys=("dX_dp", "dY_dp", "dZ_dp")):
        return SparseGradientVectorMixin._extend_gradient_vectors(results, m, n, keys)



from __future__ import absolute_import, division, print_function

import os
import shutil
import sys
import traceback

# This file needs to remain Python 2.7 compatible
# due to the underlying cctbx installer logic


installer_path = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))
libtbx_path = os.path.join(installer_path, "lib")
if libtbx_path not in sys.path:
    sys.path.append(libtbx_path)
from libtbx.auto_build import install_distribution
from libtbx.auto_build import installer_utils


class installer(install_distribution.installer):
    organization = "dials"
    product_name = "DIALS"
    dest_dir_prefix = "dials"
    make_apps = []
    configure_modules = ["dials", "xia2", "iota", "prime"]
    include_gui_packages = True
    base_package_options = ["--dials"]
    installer_dir = installer_path
    modules = [
        # hot
        "annlib",
        "ccp4io",
        # base
        "cbflib",
        "cctbx_project",
        "gui_resources",
        "ccp4io_adaptbx",
        "annlib_adaptbx",
        # dials
        "dxtbx",
        "dials",
        "xia2",
        "iota",
        "prime",
    ]
    flags = list(install_distribution.installer.flags)
    try:
        flags.remove("create_versioned_dispatchers")
    except ValueError:
        pass

    def product_specific_preinstallation_hook(self):
        prefix = os.path.abspath(self.options.prefix)
        if prefix.startswith(installer_path):
            sys.exit(
                "Invalid installation option: --prefix={givenprefix}\n\n"
                "Please install DIALS to a location outside of the installer directory.\n"
                "Suggested alternative: --prefix={suggestedprefix}".format(
                    givenprefix=self.options.prefix,
                    suggestedprefix=os.path.dirname(prefix),
                )
            )

    def reconfigure(self, log=None, *args, **kwargs):
        """Intercept any errors and print log excerpt"""
        args = list(args) + ["--skip_phenix_dispatchers"]
        try:
            # Deliberately call our alternative rather than the super()
            self.reconfigure_as_libtbx_does(log)
        except Exception:
            if not self.options.verbose:
                print("\n" + " -=-" * 20)
                print("\nAn error occurred during installation\n")
                print("Excerpt from installation log:")
                with open(log.name, "r") as fh:
                    for line in fh.readlines()[-30:]:
                        print(" :", line, end="")
                print("\nThis led to ", end="")
                sys.stdout.flush()
            traceback.print_exc()
            print("\n")
            sys.exit(
                "Please report this installation error to dials-support@lists.sourceforge.net"
            )

    def reconfigure_as_libtbx_does(self, log):
        """
        Run libtbx/configure.py to configure the build in the new location.

        This is copied from the cctbx_project/libtbx/auto_build/install_distribution.py,
        which does not allow customizing the call to configure. We want
        to customize to remove phenix dispatchers.
        """
        os.chdir(self.build_dir)

        if "darwin" in sys.platform:
            base_python = os.path.join(
                self.base_dir, "python.app", "Contents", "MacOS", "python"
            )
        elif "win32" in sys.platform:
            base_python = os.path.join(self.base_dir, "python.exe")
        else:
            base_python = os.path.join(self.base_dir, "bin", "python")

        if "win32" == sys.platform:
            args = [
                base_python,
                os.path.join(
                    self.modules_dir, "cctbx_project", "libtbx", "configure.py"
                ),
            ]
        else:
            args = [
                base_python,
                os.path.join(
                    self.modules_dir, "cctbx_project", "libtbx", "configure.py"
                ),
                "--current_working_directory",
                self.build_dir,
            ]

        args += ["--use_conda", "--skip_phenix_dispatchers"]
        args += self.configure_modules

        installer_utils.call(args=args, log=log, verbose=self.options.verbose)

    def product_specific_prepackage_hook(self, directory):
        """
        Remove irrelevant files from installer.
        """
        self.print_header("Deflating installer")

        suffixes = ["B", "KB", "MB", "GB", "TB", "PB"]

        def humansize(nbytes):
            if nbytes == 0:
                return "0 B"
            i = 0
            while nbytes >= 1024 and i < len(suffixes) - 1:
                nbytes /= 1024.0
                i += 1
            f = ("%.2f" % nbytes).rstrip("0").rstrip(".")
            return "%s %s" % (f, suffixes[i])

        self._cleaned_size, self._cleaned_files = 0, 0

        def rmdir(subdir):
            fullpath = os.path.join(directory, subdir)
            if not os.path.exists(fullpath):
                print("Skipping", " " * 26, subdir)
                return
            num_files, total_size = 0, 0
            for dirpath, dirnames, filenames in os.walk(fullpath):
                for f in filenames:
                    fp = os.path.join(dirpath, f)
                    total_size += os.path.getsize(fp)
                    num_files += 1
            print(
                "Removing %9s, %4d files from %s"
                % (humansize(total_size), num_files, subdir)
            )
            shutil.rmtree(fullpath)
            self._cleaned_size = self._cleaned_size + total_size
            self._cleaned_files = self._cleaned_files + num_files

        def rmext(subdir, extension):
            fullpath = os.path.join(directory, subdir)
            if not os.path.exists(fullpath):
                print("Skipping *%s" % extension, " " * 26, subdir)
                return
            filelist, total_size = [], 0
            for dirpath, dirnames, filenames in os.walk(fullpath):
                for f in filenames:
                    if f.endswith(extension):
                        fp = os.path.join(dirpath, f)
                        filelist.append(fp)
                        total_size += os.path.getsize(fp)
            print(
                "Removing %9s, %4d %s files from %s"
                % (humansize(total_size), len(filelist), extension, subdir)
            )
            for f in filelist:
                os.remove(f)
            self._cleaned_size = self._cleaned_size + total_size
            self._cleaned_files = self._cleaned_files + len(filelist)

        def rmfile(filename):
            fullpath = os.path.join(directory, filename)
            if not os.path.exists(fullpath):
                print("Skipping", " " * 26, filename)
                return
            filesize = os.path.getsize(fullpath)
            print("Removing %9s, file %s" % (humansize(filesize), filename))
            os.remove(fullpath)
            self._cleaned_size = self._cleaned_size + filesize
            self._cleaned_files = self._cleaned_files + 1

        # Deduce matplotlib path
        # (base/lib/python2.??/site-packages/matplotlib-????/matplotlib)
        # (base/Python.framework/Versions/?.?/lib/python?.?/site-packages/matplotlib-(...) on MacOS)
        try:
            import inspect

            import matplotlib

            matplotpath = os.path.dirname(
                os.path.dirname(inspect.getsourcefile(matplotlib))
            )
            relpath = []
            matplotpath, d = os.path.split(matplotpath)
            relpath.append(d)
            while d and (d != "base"):
                matplotpath, d = os.path.split(matplotpath)
                relpath.append(d)
            if d == "base":
                relpath.reverse()
                # delete matplotlib tests
                matplotpath = os.path.join(*relpath)
                rmdir(os.path.join(matplotpath, "matplotlib", "tests"))
                rmdir(os.path.join(matplotpath, "mpl_toolkits", "tests"))

                # ...while we're here
                sitepath = os.path.dirname(matplotpath)
                rmdir(os.path.join(sitepath, "numpy/core/tests"))
                rmdir(os.path.join(sitepath, "numpy/doc"))
                rmdir(os.path.join(sitepath, "numpy/distutils/tests"))
                rmdir(os.path.join(sitepath, "numpy/f2py/docs"))

                pythonpath = os.path.dirname(sitepath)
                rmdir(os.path.join(pythonpath, "test"))
        except Exception:
            print("Could not deduce python package paths")

        rmdir("base/man")
        rmdir("base/share/doc")
        rmdir("base/share/gtk-doc")
        rmdir("base/share/hdf5_examples")
        rmdir("base/share/man")
        rmdir("build/dials_data")
        rmdir("build/precommitbx")
        rmdir("build/regression_data")
        rmdir("build/xia2_regression")
        rmext("build", ".o")
        for f in ("setpaths", "setpaths_debug", "setpaths_all", "unsetpaths"):
            for ext in (".sh", ".csh"):
                rmfile(os.path.join("build", f + ext))
        for p in (
            "chrono",
            "date_time",
            "detail",
            "filesystem",
            "program_options",
            "python",
            "system",
            "thread",
            "timer",
        ):
            rmdir(os.path.join("modules/boost/libs", p, "example"))
            rmdir(os.path.join("modules/boost/libs", p, "doc"))
            rmdir(os.path.join("modules/boost/libs", p, "test"))
            rmdir(os.path.join("modules/boost/libs", p, "tutorial"))
        rmdir("modules/boost/libs/date_time/xmldoc")
        rmdir("modules/cbflib/doc")
        rmdir("modules/cbflib/examples")
        rmdir("modules/cbflib/ply-3.2/doc")
        rmdir("modules/cbflib/ply-3.2/example")
        rmdir("modules/cbflib/ply-3.2/test")
        rmext("modules/cbflib", ".cbf")
        rmdir("modules/clipper/examples")
        print("-" * 60)
        print(
            "Deleted %d files, decrufting installation by %s\n"
            % (self._cleaned_files, humansize(self._cleaned_size))
        )

    def check_directories(self):
        # Work out the target to be created by libtbx and check if it exists
        expected_dir = os.path.abspath(
            os.path.join(
                self.options.prefix, "%s-%s" % (self.dest_dir_prefix, self.version)
            )
        )
        regular_exists = os.path.exists(expected_dir)

        super(installer, self).check_directories()

        if self.options.raw_prefix:
            # Make sure we clean up the extra directory if created
            assert expected_dir == self.dest_dir
            if not regular_exists and os.path.exists(self.dest_dir):
                os.remove(self.dest_dir)

            # Now, update all the target paths
            self.dest_dir = self.options.prefix
            self.build_dir = os.path.join(self.dest_dir, "build")
            self.base_dir = os.path.join(self.options.prefix, "conda_base")
            self.modules_dir = os.path.join(self.options.prefix, "modules")
            os.environ[self.product_name + "_LOC"] = self.dest_dir
            os.environ[self.product_name + "_BUILD"] = self.build_dir

    def add_product_specific_options(self, parser):
        parser.add_option(
            "--raw-prefix",
            action="store_true",
            default=False,
            help="Use --prefix as the direct destination, rather than adding a dials-* subdir",
        )


if __name__ == "__main__":
    installer(sys.argv[1:]).install()



from __future__ import annotations

import logging
import math

from scitbx.matrix import col, sqr

from dials.algorithms.indexing import DialsIndexError
from dials.array_family import flex

logger = logging.getLogger(__name__)
"""
Class to determine mosaicity and effective domain size for a crystal given a set of indexed reflections
"""


class NaveParameters:
    def __init__(self, params, experiments, reflections, refinery, graph_verbose=True):
        self.params = params
        self.experiments = experiments
        self.reflections = reflections
        self.refinery = refinery
        self.graph_verbose = graph_verbose

    def __call__(self):
        """Determine optimal mosaicity and domain size model (monochromatic)"""
        if self.refinery is None:
            RR = self.reflections
        else:
            RR = self.refinery.predict_for_reflection_table(self.reflections)

        all_crystals = []
        self.nv_acceptance_flags = flex.bool(len(self.reflections["id"]))
        from dxtbx.model import MosaicCrystalSauter2014

        for iid, experiment in enumerate(self.experiments):
            excursion_rad = RR["delpsical.rad"].select(RR["id"] == iid)
            delta_psi_deg = excursion_rad * 180.0 / math.pi
            logger.info("")
            logger.info("%s %s", flex.max(delta_psi_deg), flex.min(delta_psi_deg))
            mean_excursion = flex.mean(delta_psi_deg)
            logger.info(
                "The mean excursion is %7.3f degrees, r.m.s.d %7.3f",
                mean_excursion,
                math.sqrt(flex.mean(RR["delpsical2"].select(RR["id"] == iid))),
            )

            crystal = MosaicCrystalSauter2014(self.experiments[iid].crystal)
            self.experiments[iid].crystal = crystal
            beam = self.experiments[iid].beam
            miller_indices = self.reflections["miller_index"].select(
                self.reflections["id"] == iid
            )

            # FIXME XXX revise this formula so as to use a different wavelength potentially for each reflection
            two_thetas = crystal.get_unit_cell().two_theta(
                miller_indices, beam.get_wavelength(), deg=True
            )
            dspacings = crystal.get_unit_cell().d(miller_indices)

            # First -- try to get a reasonable envelope for the observed excursions.
            # minimum of three regions; maximum of 50 measurements in each bin
            logger.info("fitting parameters on %d spots", len(excursion_rad))
            n_bins = min(max(3, len(excursion_rad) // 25), 50)
            bin_sz = len(excursion_rad) // n_bins
            logger.info("nbins %s bin_sz %s", n_bins, bin_sz)
            order = flex.sort_permutation(two_thetas)
            two_thetas_env = flex.double()
            dspacings_env = flex.double()
            excursion_rads_env = flex.double()
            for x in range(0, n_bins):
                subset = order[x * bin_sz : (x + 1) * bin_sz]
                two_thetas_env.append(flex.mean(two_thetas.select(subset)))
                dspacings_env.append(flex.mean(dspacings.select(subset)))
                excursion_rads_env.append(
                    flex.max(flex.abs(excursion_rad.select(subset)))
                )

            # Second -- parameter fit
            # solve the normal equations
            sum_inv_u_sq = flex.sum(dspacings_env * dspacings_env)
            sum_inv_u = flex.sum(dspacings_env)
            sum_te_u = flex.sum(dspacings_env * excursion_rads_env)
            sum_te = flex.sum(excursion_rads_env)
            Normal_Mat = sqr((sum_inv_u_sq, sum_inv_u, sum_inv_u, len(dspacings_env)))
            Vector = col((sum_te_u, sum_te))
            solution = Normal_Mat.inverse() * Vector
            s_ang = 1.0 / (2 * solution[0])
            logger.info("Best LSQ fit Scheerer domain size is %9.2f ang", s_ang)

            k_degrees = solution[1] * 180.0 / math.pi
            logger.info(
                "The LSQ full mosaicity is %8.5f deg; half-mosaicity %9.5f",
                2 * k_degrees,
                k_degrees,
            )

            from xfel.mono_simulation.max_like import minimizer

            # coerce the estimates to be positive for max-likelihood
            lower_limit_domain_size = (
                math.pow(crystal.get_unit_cell().volume(), 1.0 / 3.0) * 3
            )  # params.refinement.domain_size_lower_limit

            d_estimate = max(s_ang, lower_limit_domain_size)
            M = minimizer(
                d_i=dspacings,
                psi_i=excursion_rad,
                eta_rad=abs(2.0 * solution[1]),
                Deff=d_estimate,
            )
            logger.info(
                "ML: mosaicity FW=%4.2f deg, Dsize=%5.0fA on %d spots",
                M.x[1] * 180.0 / math.pi,
                2.0 / M.x[0],
                len(two_thetas),
            )
            tan_phi_rad_ML = dspacings / (2.0 / M.x[0])
            tan_phi_deg_ML = tan_phi_rad_ML * 180.0 / math.pi
            tan_outer_deg_ML = tan_phi_deg_ML + 0.5 * M.x[1] * 180.0 / math.pi

            # Only set the flags for those reflections that were indexed for this lattice
            self.nv_acceptance_flags.set_selected(
                self.reflections["id"] == iid,
                flex.abs(delta_psi_deg) < tan_outer_deg_ML,
            )

            if (
                self.graph_verbose
            ):  # params.refinement.mosaic.enable_AD14F7B: # Excursion vs resolution fit
                AD1TF7B_MAX2T = 30.0
                AD1TF7B_MAXDP = 1.0
                from matplotlib import pyplot as plt

                plt.plot(two_thetas, delta_psi_deg, "bo")
                minplot = flex.min(two_thetas)
                plt.plot([0, minplot], [mean_excursion, mean_excursion], "k-")
                LR = flex.linear_regression(two_thetas, delta_psi_deg)
                model_y = LR.slope() * two_thetas + LR.y_intercept()
                plt.plot(two_thetas, model_y, "k-")

                plt.title(
                    "ML: mosaicity FW=%4.2f deg, Dsize=%5.0fA on %d spots"
                    % (M.x[1] * 180.0 / math.pi, 2.0 / M.x[0], len(two_thetas))
                )
                plt.plot(two_thetas, tan_phi_deg_ML, "r.")
                plt.plot(two_thetas, -tan_phi_deg_ML, "r.")
                plt.plot(two_thetas, tan_outer_deg_ML, "g.")
                plt.plot(two_thetas, -tan_outer_deg_ML, "g.")
                plt.xlim([0, AD1TF7B_MAX2T])
                plt.ylim([-AD1TF7B_MAXDP, AD1TF7B_MAXDP])
                plt.show()
                plt.close()

            from xfel.mono_simulation.util import green_curve_area

            self.green_curve_area = green_curve_area(two_thetas, tan_outer_deg_ML)
            logger.info("The green curve area is %s", self.green_curve_area)

            crystal.set_half_mosaicity_deg(M.x[1] * 180.0 / (2.0 * math.pi))
            crystal.set_domain_size_ang(2.0 / M.x[0])
            self._ML_full_mosaicity_rad = M.x[1]
            self._ML_domain_size_ang = 2.0 / M.x[0]

            # params.refinement.mosaic.model_expansion_factor
            """The expansion factor should be initially set to 1, then expanded so that the # reflections matched becomes
        as close as possible to # of observed reflections input, in the last integration call.  Determine this by
        inspecting the output log file interactively.  Do not exceed the bare minimum threshold needed.
        The intention is to find an optimal value, global for a given dataset."""
            model_expansion_factor = 1.4
            crystal.set_half_mosaicity_deg(
                crystal.get_half_mosaicity_deg() * model_expansion_factor
            )
            crystal.set_domain_size_ang(
                crystal.get_domain_size_ang() / model_expansion_factor
            )

            if (
                self.ewald_proximal_volume(iid)
                > self.params.indexing.stills.ewald_proximal_volume_max
            ):
                raise DialsIndexError(
                    "Ewald proximity volume too high, %f"
                    % self.ewald_proximal_volume(iid)
                )

            all_crystals.append(crystal)
        return all_crystals

    def ewald_proximal_volume(self, expt_id=0):
        """computes the volume of reciprocal space (actually, half the volume, in this implementation) in which
        reciprocal lattice centroids will fall under the green curve.  In other words, this is proportional to the
        number of predicted reflections."""

        R_L = (
            1.0 / self.experiments[expt_id].beam.get_wavelength()
        )  # radius of Ewald sphere

        # TT is the outermost two-theta angle to perform the volume integration (hi-resolution cutoff)
        TT = 2.0 * math.asin(
            self.experiments[expt_id].beam.get_wavelength()
            / (2.0 * self.params.indexing.stills.ewald_proximity_resolution_cutoff)
        )

        part_vol = math.pi * (2.0 / 3.0) * (1.0 - math.cos(TT))
        Ewald_sphere_volume = part_vol * math.pow(
            R_L, 3.0
        )  # base volume of Ewald sphere segment
        R_prime = R_L + 1.0 / self._ML_domain_size_ang
        domain_size_volume = part_vol * math.pow(
            R_prime, 3.0
        )  # expanded volume accommodating spot size

        # complicated integral for mosaic spread volume, must be calculated numerically
        summation = 0.0
        N_terms = 100
        for x in range(N_terms):
            phi = (x / N_terms) * TT
            # inner integral over radius r
            integral = math.pow(
                R_prime + (self._ML_full_mosaicity_rad * R_L * math.sin(phi) / 2.0), 3.0
            ) - math.pow(R_prime, 3.0)
            summation += (integral * math.sin(phi)) * (TT / N_terms)
        mosaicity_volume = (2.0 / 3.0) * math.pi * summation

        return (domain_size_volume - Ewald_sphere_volume) + mosaicity_volume



from __future__ import annotations

import os

import procrunner


def test_rl_png_imported_experiments(dials_data, tmp_path):
    data_dir = dials_data("centroid_test_data", pathlib=True)
    experiments_path = data_dir / "imported_experiments.json"
    strong_pickle = data_dir / "strong.pickle"

    result = procrunner.run(
        ["dials.rl_png", experiments_path, strong_pickle], working_directory=tmp_path
    )
    assert not result.returncode and not result.stderr

    for s in (
        "beam_vector",
        "e3",
        "rotation_axis",
        "solution_1",
        "solution_2",
        "solution_3",
    ):
        assert tmp_path.joinpath(f"rl_{s}.png").is_file()


def test_rl_png_refinement_data(dials_regression, tmp_path):
    data_dir = os.path.join(dials_regression, "refinement_test_data", "i04_weak_data")
    experiments_path = os.path.join(data_dir, "experiments.json")
    indexed_pickle = os.path.join(data_dir, "indexed_strong.pickle")

    result = procrunner.run(
        ["dials.rl_png", experiments_path, indexed_pickle], working_directory=tmp_path
    )
    assert not result.returncode and not result.stderr

    for s in ("beam_vector", "e3", "rotation_axis", "a", "b", "c"):
        assert tmp_path.joinpath(f"rl_{s}.png").is_file()



from __future__ import annotations

import copy

from dxtbx.model import Experiment, ExperimentList, Scan
from dxtbx.model.experiment_list import ExperimentListFactory

from dials.array_family import flex
from dials.util.slice import slice_experiments, slice_reflections


def test_slice_experiments():
    image_range = (0, 1000)
    oscillation = (0, 0.1)
    scan = Scan(image_range, oscillation)
    experiments = ExperimentList([Experiment(scan=scan)])
    sliced_image_range = [(1, 5)]
    sliced_experiments = slice_experiments(experiments, sliced_image_range)
    assert sliced_experiments[0].scan.get_image_range() == sliced_image_range[0]
    copy.deepcopy(sliced_experiments)


def test_slice_experiments_centroid_test_data(dials_data):
    files = sorted(dials_data("centroid_test_data", pathlib=True).glob("*.cbf"))
    experiments = ExperimentListFactory.from_filenames(files)
    sliced_image_range = [(1, 3)]
    sliced_experiments = slice_experiments(experiments, sliced_image_range)
    assert sliced_experiments[0].scan.get_image_range() == sliced_image_range[0]
    # for some reason the sliced_experiments is not copyable
    assert copy.deepcopy(sliced_experiments)


def test_slice_experiments_centroid_test_data_starting_from_2(dials_data):
    files = sorted(dials_data("centroid_test_data", pathlib=True).glob("*.cbf"))[1:]
    experiments = ExperimentListFactory.from_filenames(files)
    sliced_image_range = [(2, 4)]
    sliced_experiments = slice_experiments(experiments, sliced_image_range)
    assert sliced_experiments[0].scan.get_image_range() == sliced_image_range[0]


def test_slice_reflections():
    r = flex.reflection_table()
    r["id"] = flex.int([0, 0, 0, 1, 1, 1, 2, 2, 2])
    image_number = [0, 1, 2, 0, 1, 2, 0, 1, 2]
    r["xyzobs.px.value"] = flex.vec3_double(
        zip([0] * len(r), [0] * len(r), image_number)
    )
    sliced_r = slice_reflections(r, [(1, 2), (1, 1), (2, 3)])
    assert list(sliced_r["id"]) == [0, 0, 1, 2, 2]
    assert list(sliced_r["xyzobs.px.value"].parts()[2]) == [0.0, 1.0, 0.0, 1.0, 2.0]



from __future__ import annotations

import os

from dxtbx.model import Crystal
from rstbx.cftbx.coordinate_frame_helpers import align_reference_frame
from scitbx import matrix


def dump(experiments, directory):
    """
    Dump the experiments in mosflm format

    :param experiments: The experiments to dump
    :param directory: The directory to write to
    """
    for i, experiment in enumerate(experiments):
        suffix = ""
        if len(experiments) > 1:
            suffix = "_%i" % (i + 1)

        sub_dir = f"{directory}{suffix}"
        if not os.path.isdir(sub_dir):
            os.makedirs(sub_dir)
        detector = experiment.detector
        beam = experiment.beam
        goniometer = experiment.goniometer

        # XXX imageset is getting the experimental geometry from the image files
        # rather than the input models.expt file
        imageset = experiment.imageset

        R_to_mosflm = align_reference_frame(
            beam.get_s0(),
            (1.0, 0.0, 0.0),
            goniometer.get_rotation_axis(),
            (0.0, 0.0, 1.0),
        )

        cryst = experiment.crystal
        cryst = cryst.change_basis(
            cryst.get_space_group().info().change_of_basis_op_to_reference_setting()
        )
        A = matrix.sqr(cryst.get_A())
        A_inv = A.inverse()

        real_space_a = R_to_mosflm * A_inv.elems[:3]
        real_space_b = R_to_mosflm * A_inv.elems[3:6]
        real_space_c = R_to_mosflm * A_inv.elems[6:9]

        cryst_mosflm = Crystal(
            real_space_a,
            real_space_b,
            real_space_c,
            space_group=cryst.get_space_group(),
        )
        A_mosflm = matrix.sqr(cryst_mosflm.get_A())
        U_mosflm = matrix.sqr(cryst_mosflm.get_U())
        assert U_mosflm.is_r3_rotation_matrix(), U_mosflm
        w = beam.get_wavelength()

        index_mat = os.path.join(sub_dir, "index.mat")
        mosflm_in = os.path.join(sub_dir, "mosflm.in")
        print(f"Exporting experiment to {index_mat} and {mosflm_in}")

        with open(index_mat, "w") as f:
            f.write(format_mosflm_mat(w * A_mosflm, U_mosflm, cryst.get_unit_cell()))

        img_dir, template = os.path.split(imageset.get_template())
        symmetry = cryst_mosflm.get_space_group().type().number()
        beam_centre = tuple(reversed(detector[0].get_beam_centre(beam.get_s0())))
        distance = detector[0].get_directed_distance()

        with open(mosflm_in, "w") as f:
            f.write(
                write_mosflm_input(
                    directory=img_dir,
                    template=template,
                    symmetry=symmetry,
                    beam_centre=beam_centre,
                    distance=distance,
                    mat_file="index.mat",
                )
            )


def format_mosflm_mat(A, U, unit_cell, missets=(0, 0, 0)):
    lines = []
    uc_params = unit_cell.parameters()
    for i in range(3):
        lines.append(("%12.8f" * 3) % A.elems[i * 3 : 3 * (i + 1)])
    lines.append(("%12.3f" * 3) % missets)
    for i in range(3):
        lines.append("%12.8f" * 3 % U.elems[i * 3 : 3 * (i + 1)])
    lines.append(("%12.4f" * 6) % uc_params)
    lines.append(("%12.3f" * 3) % missets)
    return "\n".join(lines)


def write_mosflm_input(
    directory=None,
    template=None,
    symmetry=None,
    beam_centre=None,
    distance=None,
    mat_file=None,
):
    lines = []
    if directory is not None:
        lines.append(f"DIRECTORY {directory}")
    if template is not None:
        lines.append(f"TEMPLATE {template}")
    if symmetry is not None:
        lines.append(f"SYMMETRY {symmetry}")
    if beam_centre is not None:
        lines.append("BEAM %.3f %.3f" % beam_centre)
    if distance is not None:
        lines.append(f"DISTANCE {distance:.4f}")
    if mat_file is not None:
        lines.append(f"MATRIX {mat_file}")
    return "\n".join(lines)



from __future__ import annotations

from cctbx import sgtbx, uctbx
from scitbx.array_family import flex

from dials.algorithms.integration import filtering


def test_powder_ring_filter():
    unit_cell = uctbx.unit_cell((4.498, 4.498, 7.338, 90, 90, 120))
    space_group = sgtbx.space_group_info(number=194).group()
    d_spacings = flex.double([1.9220704466392748])
    d_min = flex.min(d_spacings)
    ice_filter = filtering.PowderRingFilter(unit_cell, space_group, d_min, width=0.004)
    assert min(uctbx.d_star_sq_as_d(ice_filter.d_star_sq)) < d_min
    assert all(ice_filter(d_spacings))



"""
Testing for other refiner/refinement small utilities that can be tested
alone.
"""


from __future__ import annotations

from unittest.mock import Mock, patch

from dials.algorithms.refinement.refiner import _copy_experiments_for_refining


def test_that_the_experiment_reduced_copier_works_as_intended():
    sample = Mock()
    with patch("dials.algorithms.refinement.refiner.ExperimentList", new=list):
        dupe = _copy_experiments_for_refining(sample)[0]
    assert dupe is not sample
    # Make sure that anything refined is unique
    for att in ["beam", "goniometer", "detector", "crystal"]:
        assert getattr(sample, att) is not getattr(dupe, att)
    # Anything read-only should be untouched
    for att in ["scan", "profile", "imageset", "scaling_model"]:
        assert getattr(sample, att) is getattr(dupe, att)



"""
Test for the target function module.
"""

from __future__ import annotations

import copy
from unittest.mock import MagicMock, Mock, patch

import numpy as np
import pytest
from scipy.sparse import csc_matrix

from dxtbx.model import Crystal, Experiment, Scan
from libtbx import phil
from scitbx import sparse

from dials.algorithms.scaling.active_parameter_managers import (
    multi_active_parameter_manager,
)
from dials.algorithms.scaling.Ih_table import IhTable
from dials.algorithms.scaling.model.model import PhysicalScalingModel
from dials.algorithms.scaling.parameter_handler import scaling_active_parameter_manager
from dials.algorithms.scaling.scaler import SingleScaler
from dials.algorithms.scaling.target_function import ScalingTarget, ScalingTargetFixedIH
from dials.array_family import flex
from dials.util.options import ArgumentParser


@pytest.fixture
def large_reflection_table():
    """Generate reflection table to test the basis and target function."""
    # these miller_idx/d_values don't make physical sense, but I didn't want to
    # have to write the tests for lots of reflections.
    reflections = flex.reflection_table()
    reflections["intensity"] = flex.double(
        [75.0, 10.0, 100.0, 25.0, 50.0, 100.0, 25.0, 20.0, 300.0, 10.0]
    )
    reflections["variance"] = flex.double(
        [50.0, 10.0, 100.0, 50.0, 10.0, 100.0, 50.0, 10.0, 100.0, 10.0]
    )
    reflections["inverse_scale_factor"] = flex.double(10, 1.0)
    reflections["miller_index"] = flex.miller_index(
        [
            (1, 0, 0),
            (0, 0, 1),
            (1, 0, 0),
            (1, 0, 0),
            (0, 0, 1),
            (1, 0, 0),
            (0, 4, 0),
            (0, 0, 1),
            (1, 0, 0),
            (0, 4, 0),
        ]
    )  # don't change
    reflections["d"] = flex.double(
        [2.0, 0.8, 2.0, 2.0, 0.8, 2.0, 2.0, 0.8, 2.0, 1.0]
    )  # don't change
    reflections["partiality"] = flex.double(10, 1.0)
    reflections["xyzobs.px.value"] = flex.vec3_double(
        [(0.0, 0.0, i) for i in range(0, 45, 5)] + [(0.0, 0.0, 59.0)]
    )
    reflections["s1"] = flex.vec3_double([(0.0, 0.1, 1.0)] * 2 + [(0.0, 0.1, 20.0)] * 8)
    return reflections


@pytest.fixture
def small_reflection_table():
    """Generate reflection table to test the basis and target function."""
    # these miller_idx/d_values don't make physical sense, but I didn't want to
    # have to write the tests for lots of reflections.
    reflections = flex.reflection_table()
    reflections["intensity"] = flex.double([75.0, 10.0, 100.0, 50.0, 65.0])
    reflections["variance"] = flex.double([50.0, 10.0, 100.0, 50.0, 65.0])
    reflections["inverse_scale_factor"] = flex.double(5, 1.0)
    reflections["miller_index"] = flex.miller_index(
        [(1, 0, 0), (0, 0, 1), (1, 0, 0), (0, 0, 1), (0, 0, 2)]
    )  # don't change
    reflections["d"] = flex.double([2.0, 0.8, 2.0, 0.8, 1.5])  # don't change
    reflections["xyzobs.px.value"] = flex.vec3_double(
        [(0.0, 0.0, i) for i in [0.0, 5.0, 10.0, 2.0, 8.0]]
    )
    reflections["s1"] = flex.vec3_double([(0.0, 0.1, 1.0)] * 2 + [(0.0, 0.1, 20.0)] * 3)
    return reflections


@pytest.fixture
def single_exp():
    """Generate an experiment object."""
    crystal = Crystal.from_dict(
        {
            "__id__": "crystal",
            "real_space_a": [1.0, 0.0, 0.0],
            "real_space_b": [0.0, 1.0, 0.0],
            "real_space_c": [0.0, 0.0, 2.0],
            "space_group_hall_symbol": " C 2y",
        }
    )
    scan = Scan(image_range=[0, 60], oscillation=[0.0, 1.0])
    return Experiment(scan=scan, crystal=crystal)


@pytest.fixture
def physical_param():
    """Generate the scaling phil param scope."""
    phil_scope = phil.parse(
        """
      include scope dials.algorithms.scaling.model.model.model_phil_scope
      include scope dials.algorithms.scaling.scaling_options.phil_scope
  """,
        process_includes=True,
    )

    parser = ArgumentParser(phil=phil_scope, check_format=False)
    parameters, _ = parser.parse_args(args=[], quick_parse=True, show_diff_phil=False)
    parameters.model = "physical"
    parameters.physical.absorption_correction = False
    return parameters


def mock_single_Ih_table():
    """Mock Ih table to use for testing the target function."""
    Ih_table = Mock()
    Ih_table.inverse_scale_factors = np.array([1.0, 1.0 / 1.1, 1.0])
    Ih_table.intensities = np.array([10.0, 10.0, 12.0])
    Ih_table.Ih_values = np.array([11.0, 11.0, 11.0])
    # These values should give residuals of [-1.0, 0.0, 1.0]
    Ih_table.weights = np.array([1.0, 1.0, 1.0])
    Ih_table.size = 3
    Ih_table.derivatives = sparse.matrix(3, 1, [{0: 1.0, 1: 2.0, 2: 3.0}])
    Ih_table.h_index_matrix = sparse.matrix(3, 2, [{0: 1, 1: 1}, {2: 1}])
    Ih_table.h_expand_matrix = Ih_table.h_index_matrix.transpose()
    Ih_table._csc_h_index_matrix = csc_matrix(
        (np.array([1, 1, 1]), (np.array([0, 1, 2]), np.array([0, 0, 1])))
    )

    def sum_in_groups_side_effect(*args):
        return args[0] * Ih_table._csc_h_index_matrix

    Ih_table.sum_in_groups = sum_in_groups_side_effect
    return Ih_table


def mock_Ih_table():
    """A mock Ih table for testing a target function."""
    Ih_table = MagicMock()
    Ih_table.blocked_data_list = [mock_single_Ih_table()]
    return Ih_table


@pytest.fixture
def mock_restrained_component():
    """Mock a component with restraints."""
    component = Mock()
    component.n_params = 3
    component.calculate_restraints.return_value = [
        flex.double([1.0, 2.0, 3.0]),
        flex.double([0.1, 0.2, 0.3]),
    ]
    jacobian_restr = sparse.matrix(component.n_params, component.n_params)
    jacobian_restr[0, 0] = 1.0
    component.calculate_jacobian_restraints.return_value = [
        flex.double([1.0, 1.1, 1.2]),
        jacobian_restr,
        flex.double([1.0, 2.0, 3.0]),
    ]
    return component


@pytest.fixture
def mock_unrestrained_component():
    """Mock a component without restraints."""
    component = Mock()
    component.n_params = 5
    component.calculate_restraints.return_value = None
    component.calculate_jacobian_restraints.return_value = None
    return component


def _component_to_apm(component):
    mock_single_apm = Mock()
    mock_single_apm.components = {
        "restrained": {
            "object": component,
            "n_params": component.n_params,
            "start_idx": 0,
        }
    }
    mock_single_apm.n_active_params = component.n_params
    return mock_single_apm


def _single_to_multi_apm(single_apm):
    apm = MagicMock()
    apm.apm_list = [single_apm]
    apm.n_active_params = single_apm.n_active_params
    apm.apm_data = {0: {"start_idx": 0, "end_idx": single_apm.n_active_params}}
    return apm


@pytest.fixture
def mock_apm_restrained(mock_restrained_component):
    """Make a parameter manager with a restrained component"""
    return _single_to_multi_apm(_component_to_apm(mock_restrained_component))


@pytest.fixture
def mock_apm_unrestrained(mock_unrestrained_component):
    """Make a parameter manager with an unrestrained component"""
    return _single_to_multi_apm(_component_to_apm(mock_unrestrained_component))


def test_target_function_methods():
    """Test for the target methods required for the refinement engine."""
    target = ScalingTarget()
    r, w = target.compute_residuals(mock_single_Ih_table())
    assert r.size() == w.size()
    assert r == pytest.approx([-1.0, 0.0, 1.0])
    assert w == pytest.approx([1.0, 1.0, 1.0])

    f, g = target.compute_functional_gradients(mock_single_Ih_table())
    assert f == pytest.approx(2.0)
    assert g == pytest.approx([-19.04072398])

    r2, j, w2 = target.compute_residuals_and_gradients(mock_single_Ih_table())
    assert r == r2
    assert w == w2
    assert j.n_cols == 1 and j.n_rows == 3


def test_target_function_restraints_methods(mock_apm_restrained, mock_apm_unrestrained):
    """Test for the target restraints methods required for the refinement engine."""
    target = ScalingTarget()

    restraints = target.compute_restraints_residuals_and_gradients(mock_apm_restrained)
    assert len(restraints) == 3
    assert target.param_restraints is True
    assert list(restraints[0]) == [1.0, 1.1, 1.2]
    assert list(restraints[2]) == [1.0, 2.0, 3.0]

    restraints = target.compute_restraints_functional_gradients(mock_apm_restrained)
    assert len(restraints) == 2
    assert restraints[0] == 6.0
    assert list(restraints[1]) == [0.1, 0.2, 0.3]

    achieved = target.achieved()
    assert isinstance(achieved, bool)

    restraints = target.compute_restraints_residuals_and_gradients(
        mock_apm_unrestrained
    )
    assert restraints is None
    assert target.param_restraints is False
    # Check that if called again, returns None without doing calculations
    restraints = target.compute_restraints_residuals_and_gradients([])
    assert restraints is None

    target = ScalingTarget()  # Need to make new instance or won't calc restr as
    # param_restraints is set to False
    assert target.param_restraints is True
    restraints = target.compute_restraints_functional_gradients(mock_apm_unrestrained)
    assert restraints is None
    assert target.param_restraints is False


def test_target_rmsd_calculation(mock_apm_restrained, mock_apm_unrestrained):
    """Test the RMSD calculation, with and without restraints."""
    target = ScalingTarget()
    assert target.param_restraints is True
    # with input, expect residuals of [-1, 0, 1], weights of [1, 1, 1],
    # restraints of [1, 2, 3], so expect residual of sqrt((2+6)/3)
    rmsds = target.rmsds(mock_Ih_table(), mock_apm_restrained)
    assert len(rmsds) == 1
    assert rmsds[0] == pytest.approx((8.0 / 3.0) ** 0.5, abs=1e-6)
    assert target.param_restraints is True

    # test rmsd calculation without restraints
    rmsds = target.rmsds(mock_Ih_table(), mock_apm_unrestrained)
    assert len(rmsds) == 1
    assert rmsds[0] == pytest.approx((2.0 / 3.0) ** 0.5, abs=1e-6)
    assert target.param_restraints is False


def test_target_fixedIh():
    """Test the target function for targeted scaling (where Ih is fixed)."""

    target = ScalingTargetFixedIH()
    Ih_table = mock_Ih_table().blocked_data_list[0]
    R, _ = target.compute_residuals(Ih_table)
    expected_residuals = flex.double([-1.0, 0.0, 1.0])
    assert list(R) == pytest.approx(list(expected_residuals))
    _, G = target.compute_functional_gradients(Ih_table)
    assert list(G) == pytest.approx([-44.0])
    # Add in finite difference check

    Ih_table = mock_Ih_table().blocked_data_list[0]
    J = target.calculate_jacobian(Ih_table)
    assert J.n_cols == 1
    assert J.n_rows == 3
    assert J.non_zeroes == 3
    assert J[0, 0] == pytest.approx(-11.0)
    assert J[1, 0] == pytest.approx(-22.0)
    assert J[2, 0] == pytest.approx(-33.0)

    expected_rmsd = (
        flex.sum(flex.pow2(expected_residuals)) / len(expected_residuals)
    ) ** 0.5
    assert target._rmsds is None
    target.param_restraints = False  # don't try to use apm to get restraints
    assert target.rmsds(mock_Ih_table(), [])
    assert target._rmsds == pytest.approx([expected_rmsd])


# For testing the targetfunction calculations using finite difference methods,
# need to initialise real instances of the scaling datastructures to allow
# variation of the parameters and updating of the linked datastructures.


def test_target_gradient_calculation_finite_difference(
    small_reflection_table, single_exp, physical_param
):
    """Test the calculated gradients against a finite difference calculation."""
    model = PhysicalScalingModel.from_data(
        physical_param, single_exp, small_reflection_table
    )

    # need to 'add_data'
    model.configure_components(small_reflection_table, single_exp, physical_param)
    model.components["scale"].update_reflection_data()
    model.components["decay"].update_reflection_data()
    apm = multi_active_parameter_manager(
        ScalingTarget(),
        [model.components],
        [["scale", "decay"]],
        scaling_active_parameter_manager,
    )
    model.components["scale"].inverse_scales = flex.double([2.0, 1.0, 2.0])
    model.components["decay"].inverse_scales = flex.double([1.0, 1.0, 0.4])

    Ih_table = IhTable([small_reflection_table], single_exp.crystal.get_space_group())

    with patch.object(SingleScaler, "__init__", lambda x, y, z, k: None):
        scaler = SingleScaler(None, None, None)
        scaler._Ih_table = Ih_table

        # Now do finite difference check.
        target = ScalingTarget()

        scaler.update_for_minimisation(apm, 0)
        grad = target.calculate_gradients(scaler.Ih_table.blocked_data_list[0])
        res, _ = target.compute_residuals(scaler.Ih_table.blocked_data_list[0])

        assert (
            res > 1e-8
        ), """residual should not be zero, or the gradient test
        below will not really be working!"""

        # Now compare to finite difference
        f_d_grad = calculate_gradient_fd(target, scaler, apm)
        print(list(f_d_grad))
        print(list(grad))
        assert list(grad) == pytest.approx(list(f_d_grad))

        sel = f_d_grad > 1e-8
        assert sel, """assert sel has some elements, as finite difference grad should
        not all be zero, or the test will not really be working!
        (expect one to be zero for KB scaling example?)"""


def test_target_jacobian_calculation_finite_difference(
    physical_param, single_exp, large_reflection_table
):
    """Test the calculated jacobian against a finite difference calculation."""
    physical_param.physical.decay_correction = False
    model = PhysicalScalingModel.from_data(
        physical_param, single_exp, large_reflection_table
    )
    # need to 'add_data'
    model.configure_components(large_reflection_table, single_exp, physical_param)
    model.components["scale"].update_reflection_data()
    apm = multi_active_parameter_manager(
        ScalingTarget(),
        [model.components],
        [["scale"]],
        scaling_active_parameter_manager,
    )
    Ih_table = IhTable([large_reflection_table], single_exp.crystal.get_space_group())

    with patch.object(SingleScaler, "__init__", lambda x, y, z, k: None):
        scaler = SingleScaler(None, None, None)
        scaler._Ih_table = Ih_table

        target = ScalingTarget()
        scaler.update_for_minimisation(apm, 0)

        fd_jacobian = calculate_jacobian_fd(target, scaler, apm)
        r, jacobian, w = target.compute_residuals_and_gradients(
            scaler.Ih_table.blocked_data_list[0]
        )
        assert r == pytest.approx(
            [-50.0 / 3.0, 70.0 / 3.0, -20.0 / 3.0, 12.5, -2.5]
            + [-25.0, 0.0, -75.0, 0.0, 200.0]
        )
        assert w == pytest.approx(
            [0.1, 0.1, 0.1, 0.02, 0.1, 0.02, 0.01, 0.02, 0.01, 0.01]
        )

        n_rows = jacobian.n_rows
        n_cols = jacobian.n_cols

        print(jacobian)
        print(fd_jacobian)

        for i in range(0, n_rows):
            for j in range(0, n_cols):
                assert jacobian[i, j] == pytest.approx(fd_jacobian[i, j], abs=1e-4)


def calculate_gradient_fd(target, scaler, apm):
    """Calculate gradient array with finite difference approach."""
    delta = 1.0e-6
    gradients = flex.double([0.0] * apm.n_active_params)
    Ih_table = scaler.Ih_table.blocked_data_list[0]
    # iterate over parameters, varying one at a time and calculating the gradient
    for i in range(apm.n_active_params):
        new_x = copy.copy(apm.x)
        new_x[i] -= 0.5 * delta
        apm.set_param_vals(new_x)
        scaler.update_for_minimisation(apm, 0)
        R_low = np.square(target.calculate_residuals(Ih_table)) * Ih_table.weights
        new_x[i] += delta
        apm.set_param_vals(new_x)
        scaler.update_for_minimisation(apm, 0)
        R_upper = np.square(target.calculate_residuals(Ih_table)) * Ih_table.weights
        new_x[i] -= 0.5 * delta
        apm.set_param_vals(new_x)
        scaler.update_for_minimisation(apm, 0)
        gradients[i] = (np.sum(R_upper) - np.sum(R_low)) / delta
    return gradients


def calculate_jacobian_fd(target, scaler, apm, block_id=0):
    """Calculate jacobian matrix with finite difference approach."""
    delta = 1.0e-7
    jacobian = sparse.matrix(
        scaler.Ih_table.blocked_data_list[block_id].size, apm.n_active_params
    )
    Ih_table = scaler.Ih_table.blocked_data_list[block_id]
    # iterate over parameters, varying one at a time and calculating the residuals
    for i in range(apm.n_active_params):
        new_x = copy.copy(apm.x)
        new_x[i] -= 0.5 * delta
        apm.set_param_vals(new_x)
        scaler.update_for_minimisation(apm, 0)
        R_low = target.calculate_residuals(Ih_table)  # unweighted unsquared residual
        new_x[i] += delta
        apm.set_param_vals(new_x)
        scaler.update_for_minimisation(apm, 0)
        R_upper = target.calculate_residuals(Ih_table)  # unweighted unsquared residual
        new_x[i] -= 0.5 * delta
        apm.set_param_vals(new_x)
        scaler.update_for_minimisation(apm, 0)
        fin_difference = (R_upper - R_low) / delta
        for j in range(fin_difference.size):
            jacobian[j, i] = fin_difference[j]
    return jacobian



from __future__ import annotations

import iotbx.phil
from cctbx.array_family import flex

import dials.algorithms.indexing.compare_orientation_matrices
import dials.util
from dials.util.options import ArgumentParser, flatten_experiments

help_message = """

Computes the change of basis operator that minimises the difference between
two orientation matrices, and calculates the rotation matrix and Euler angles
that relate the two resulting orientation matrices (after transformation by
the calculated change of basis operator). Optionally calculates the angle
between given miller indices for the respective orientation matrices.

Examples::

  dials.compare_orientation_matrices models.expt

  dials.compare_orientation_matrices models_1.expt models_2.expt

  dials.compare_orientation_matrices models_1.expt models_2.expt hkl=1,0,0
"""


phil_scope = iotbx.phil.parse(
    """
hkl = None
  .type = ints(size=3)
  .multiple=True
comparison = *pairwise sequential
  .type = choice
space_group = None
  .type = space_group
"""
)


@dials.util.show_mail_handle_errors()
def run(args=None):
    usage = "dials.compare_orientation_matrices [options] models.expt"

    parser = ArgumentParser(
        usage=usage,
        phil=phil_scope,
        read_experiments=True,
        check_format=False,
        epilog=help_message,
    )

    params, options = parser.parse_args(args, show_diff_phil=True)
    experiments = flatten_experiments(params.input.experiments)
    if len(experiments) <= 1:
        parser.print_help()
        return

    hkl = flex.miller_index(params.hkl)

    crystals = []
    for experiment in experiments:
        crystal = experiment.crystal
        if params.space_group is not None:
            crystal.set_space_group(params.space_group.group())
        crystals.append(crystal)

    rmd = dials.algorithms.indexing.compare_orientation_matrices.rotation_matrix_differences(
        crystals, miller_indices=hkl, comparison=params.comparison
    )

    print(rmd)


if __name__ == "__main__":
    run()



from __future__ import annotations

import math
import random

from scitbx.array_family import flex

from dials.algorithms.polygon import clip


def generate_intersecting(subject_size=None, target_size=4):
    if subject_size is None:
        subject_size = random.randint(3, 10)

    bbox = (0.0, 10.0, 0.0, 10.0)
    subject = generate_polygon(subject_size, bbox)
    clip = generate_polygon(target_size, bbox)

    return subject, clip


def generate_non_intersecting(subject_size=None, target_size=4):
    if subject_size is None:
        subject_size = random.randint(3, 10)

    bbox = (0.0, 10.0, 0.0, 10.0)
    subject = generate_polygon(subject_size, bbox)
    target = generate_polygon(target_size, bbox)

    offset = [
        lambda: [(x + 10.0, y) for x, y in subject],
        lambda: [(x - 10.0, y) for x, y in subject],
        lambda: [(x, y + 10.0) for x, y in subject],
        lambda: [(x, y - 10.0) for x, y in subject],
        lambda: [(x + 10.0, y - 10.0) for x, y in subject],
        lambda: [(x - 10.0, y - 10.0) for x, y in subject],
        lambda: [(x + 10.0, y + 10.0) for x, y in subject],
        lambda: [(x - 10.0, y + 10.0) for x, y in subject],
    ]

    subject = offset[random.randint(0, 7)]()
    return subject, target


def generate_polygon(nvert, box):
    xc = (box[0] + box[1]) / 2
    yc = (box[2] + box[3]) / 2
    maxr = min([xc, yc])
    minr = 0.1 * maxr
    v = []
    angle = 2 * math.pi
    dt = 2 * math.pi / nvert
    for i in range(nvert):
        r = random.uniform(minr, maxr)
        x = r * math.cos(angle)
        y = r * math.sin(angle)
        v.append((x, y))
        angle = angle + dt

    return v


def test_SimpleWithConvex_intersecting():
    for i in range(10000):

        # Generate intersecting polygons
        subject, target = generate_intersecting()

        # Do the clipping
        result = clip.simple_with_convex(
            flex.vec2_double(subject), flex.vec2_double(target)
        )

        # Ensure we have roughly valid number of vertices
        assert len(result) >= 3
        assert len(result) >= min([len(subject), len(target)])


def test_SimpleWithConvex_non_intersecting():
    for i in range(10000):

        # Generate nonintersecting polygons
        subject, target = generate_non_intersecting()

        # Do the clipping
        result = clip.simple_with_convex(
            flex.vec2_double(subject), flex.vec2_double(target)
        )

        # Ensure we no vertices
        assert len(result) == 0


def test_SimpleWithRect_intersecting():
    for i in range(10000):

        # Generate intersecting polygons
        subject, target = generate_intersecting(target_size=2)
        rect = ((0, 0), (10, 10))

        # Do the clipping
        result = clip.simple_with_rect(flex.vec2_double(subject), rect)

        # Ensure we have roughly valid number of vertices
        assert len(result) >= 3
        assert len(result) >= min([len(subject), 4])


def test_TriangleWithTriangle_intersecting():
    for i in range(10000):

        # Generate intersecting polygons
        subject, target = generate_intersecting(3, 3)

        # Do the clipping
        result = clip.triangle_with_triangle(subject, target)

        # Ensure we have roughly valid number of vertices
        assert len(result) >= 3
        assert len(result) >= min([len(subject), len(target)])


def test_TriangleWithTriangle_non_intersecting():
    for i in range(10000):

        # Generate nonintersecting polygons
        subject, target = generate_non_intersecting(3, 3)

        # Do the clipping
        result = clip.triangle_with_triangle(subject, target)

        # Ensure we no vertices
        assert len(result) == 0


def test_TriangleWithConvexQuad_intersecting():
    for i in range(10000):

        # Generate intersecting polygons
        subject, target = generate_intersecting(3, 4)

        # Do the clipping
        result = clip.triangle_with_convex_quad(subject, target)

        # Ensure we have roughly valid number of vertices
        assert len(result) >= 3
        assert len(result) >= min([len(subject), len(target)])


def test_TriangleWithConvexQuad_non_intersecting():
    for i in range(10000):

        # Generate nonintersecting polygons
        subject, target = generate_non_intersecting(3, 4)

        # Do the clipping
        result = clip.triangle_with_convex_quad(subject, target)

        # Ensure we no vertices
        assert len(result) == 0


def test_QuadWithTriangle_intersecting():
    for i in range(10000):

        # Generate intersecting polygons
        subject, target = generate_intersecting(4, 3)

        # Do the clipping
        result = clip.quad_with_triangle(subject, target)

        # Ensure we have roughly valid number of vertices
        assert len(result) >= 3
        assert len(result) >= min([len(subject), len(target)])


def test_QuadWithTriangle_non_intersecting():
    for i in range(10000):

        # Generate nonintersecting polygons
        subject, target = generate_non_intersecting(4, 3)

        # Do the clipping
        result = clip.quad_with_triangle(subject, target)

        # Ensure we no vertices
        assert len(result) == 0


def test_QuadWithConvexQuad_intersecting():
    for i in range(10000):

        # Generate intersecting polygons
        subject, target = generate_intersecting(4, 4)

        # Do the clipping
        result = clip.quad_with_convex_quad(subject, target)

        # Ensure we have roughly valid number of vertices
        assert len(result) >= 3
        assert len(result) >= min([len(subject), len(target)])


def test_QuadWithConvexQuad_non_intersecting():
    for i in range(10000):

        # Generate nonintersecting polygons
        subject, target = generate_non_intersecting(4, 4)

        # Do the clipping
        result = clip.quad_with_convex_quad(subject, target)

        # Ensure we no vertices
        assert len(result) == 0


class TestLineWithRect:
    def test(self):
        self.box = ((-10, -10), (10, 10))
        for i in range(1000):
            point1 = random.uniform(-20, 20), random.uniform(-20, 20)
            point2 = random.uniform(-20, 20), random.uniform(-20, 20)
            line = (point1, point2)
            line, status = clip.line_with_rect(line, self.box)
            if self.intersects(point1, point2):
                assert status
            else:
                assert status is False

    @staticmethod
    def inbetween(x, x0, x1):
        x00 = min([x0, x1])
        x11 = max([x0, x1])
        return x >= x00 and x <= x11

    def intersects(self, point1, point2):
        if self.is_outside(point1) and self.is_outside(point2):
            m = (point2[1] - point1[1]) / (point2[0] - point1[0])
            c = point1[1] - m * point1[0]
            x = self.box[0][0]
            y = m * x + c
            if (
                y >= self.box[0][1]
                and y <= self.box[1][1]
                and self.inbetween(y, point1[1], point2[1])
            ):
                return True
            x = self.box[1][0]
            y = m * x + c
            if (
                y >= self.box[0][1]
                and y <= self.box[1][1]
                and self.inbetween(y, point1[1], point2[1])
            ):
                return True
            y = self.box[0][1]
            x = (y - c) / m
            if (
                x >= self.box[0][0]
                and x <= self.box[1][0]
                and self.inbetween(x, point1[0], point2[0])
            ):
                return True
            y = self.box[1][1]
            x = (y - c) / m
            if (
                x >= self.box[0][0]
                and x <= self.box[0][1]
                and self.inbetween(x, point1[0], point2[0])
            ):
                return True
            return False
        else:
            return True

    def is_outside(self, point):
        return (
            point[0] < self.box[0][0]
            or point[1] < self.box[0][1]
            or point[0] > self.box[1][0]
            or point[1] > self.box[1][1]
        )



"""
Module of library functions, to perform core scaling operations on reflection
tables and experiment lists. Some functions, such as create_scaling_model and
merging statistics calculations are called from the main dials.scale script,
whereas others are provided as library functions for calling from custom
scripts. The functions defined here should ideally only require reflection
tables and ExperimentList objects (and sometimes phil_scope objects if
necessary), and return common dials objects such as reflection tables and
ExperimentLists.
"""

from __future__ import annotations

import logging
from copy import deepcopy
from unittest.mock import Mock

import numpy as np
import pkg_resources

import iotbx.merging_statistics
from cctbx import crystal, miller, uctbx
from dxtbx.model import Experiment
from dxtbx.util import ersatz_uuid4
from libtbx import Auto, phil

from dials.algorithms.scaling.Ih_table import IhTable
from dials.algorithms.scaling.model.model import KBScalingModel, PhysicalScalingModel
from dials.algorithms.scaling.scaling_utilities import (
    DialsMergingStatisticsError,
    calculate_prescaling_correction,
)
from dials.array_family import flex
from dials.util import Sorry
from dials.util.options import ArgumentParser
from dials.util.reference import intensities_from_reference_file

logger = logging.getLogger("dials")


def set_image_ranges_in_scaling_models(experiments):
    """Set the batch range in scaling models if not already set."""
    for exp in experiments:
        if exp.scan:
            valid_image_ranges = exp.scan.get_valid_image_ranges(exp.identifier)
            if "valid_image_range" not in exp.scaling_model.configdict:
                # only set if not currently set i.e. set initial
                exp.scaling_model.set_valid_image_range(exp.scan.get_image_range())
            if exp.scaling_model.configdict["valid_image_range"] != [
                valid_image_ranges[0][0],
                valid_image_ranges[-1][1],
            ]:
                # first and last values in whole list of tuples
                exp.scaling_model.limit_image_range(
                    (valid_image_ranges[0][0], valid_image_ranges[-1][1])
                )
    return experiments


def choose_initial_scaling_intensities(reflection_table, intensity_choice="profile"):
    """Choose which intensities to initially use for scaling. The LP, QE and
    partiality corrections are also applied. Two new columns are
    added to the reflection table 'intensity' and 'variance', which have
    all corrections applied except an inverse scale factor."""
    if intensity_choice == "profile":
        intensity_choice = "prf"  # rename to allow string matching with refl table
    if "intensity.prf.value" not in reflection_table:
        intensity_choice = "sum"
    elif intensity_choice == "prf":
        if (
            reflection_table.get_flags(reflection_table.flags.integrated_prf).count(
                True
            )
            == 0
        ):
            logger.warning(
                "No profile fitted reflections in this dataset, using summation intensities"
            )
            intensity_choice = "sum"
    reflection_table = calculate_prescaling_correction(reflection_table)
    conv = reflection_table["prescaling_correction"]
    # if prf/sum, use those. If combine, use prf else sum for each refl.
    if intensity_choice == "prf":
        reflection_table["intensity"] = reflection_table["intensity.prf.value"] * conv
        reflection_table["variance"] = (
            reflection_table["intensity.prf.variance"] * conv * conv
        )
    else:
        # first fill in summation intensities.
        if "partiality" in reflection_table:
            inverse_partiality = flex.double(reflection_table.size(), 1.0)
            nonzero_partiality_sel = reflection_table["partiality"] > 0.0
            good_refl = reflection_table.select(reflection_table["partiality"] > 0.0)
            inverse_partiality.set_selected(
                nonzero_partiality_sel.iselection(), 1.0 / good_refl["partiality"]
            )
            reflection_table["intensity"] = (
                reflection_table["intensity.sum.value"] * conv * inverse_partiality
            )
            reflection_table["variance"] = reflection_table[
                "intensity.sum.variance"
            ] * flex.pow2(conv * inverse_partiality)
            if "partiality.inv.variance" in reflection_table:
                reflection_table["variance"] += (
                    reflection_table["intensity.sum.value"]
                    * conv
                    * reflection_table["partiality.inv.variance"]
                )
        else:
            reflection_table["intensity"] = (
                reflection_table["intensity.sum.value"] * conv
            )
            reflection_table["variance"] = (
                reflection_table["intensity.sum.variance"] * conv * conv
            )
        if intensity_choice == "combine":
            # now overwrite prf if we have it.
            sel = reflection_table.get_flags(reflection_table.flags.integrated_prf)
            isel = sel.iselection()
            Iprf = (reflection_table["intensity.prf.value"] * conv).select(sel)
            Vprf = (reflection_table["intensity.prf.variance"] * conv * conv).select(
                sel
            )
            reflection_table["intensity"].set_selected(isel, Iprf)
            reflection_table["variance"].set_selected(isel, Vprf)
    variance_mask = reflection_table["variance"] <= 0.0
    reflection_table.set_flags(
        variance_mask, reflection_table.flags.excluded_for_scaling
    )
    return reflection_table


def scale_against_target(
    reflection_table,
    experiment,
    target_reflection_table,
    target_experiment,
    params=None,
    model="KB",
):
    """Determine scale factors for a single dataset, by scaling against a target
    reflection table. Requires a single reflection table for the reflections to
    scale and the target dataset, and an ExperimentList for both datasets. The
    params option can also be specified, if None then the default scaling
    configuration is used. The scaling model can be specified individually.

    Returns the reflection table, with added columns 'inverse_scale_factor' and
    'inverse_scale_factor_variance'."""

    if not params:
        phil_scope = phil.parse(
            """
      include scope dials.algorithms.scaling.scaling_options.phil_scope
      include scope dials.algorithms.scaling.model.model.model_phil_scope
      include scope dials.algorithms.scaling.scaling_refiner.scaling_refinery_phil_scope
    """,
            process_includes=True,
        )
        parser = ArgumentParser(phil=phil_scope, check_format=False)
        params, _ = parser.parse_args(args=[], quick_parse=True)
        params.model = model

    from dials.algorithms.scaling.scaler_factory import TargetScalerFactory

    reflections = [reflection_table, target_reflection_table]
    experiment.append(target_experiment[0])
    experiments = create_scaling_model(params, experiment, reflections)
    experiments[-1].scaling_model.set_scaling_model_as_scaled()
    scaler = TargetScalerFactory.create(params, experiments, reflections)
    scaler.perform_scaling()
    scaler.expand_scales_to_all_reflections(calc_cov=True)
    return scaler.unscaled_scalers[0].reflection_table


def scale_single_dataset(reflection_table, experiment, params=None, model="physical"):
    """Determine scale factors for a single dataset. Requires a reflection table
    and an ExperimentList with a single experiment. A custom params option can be
    specified, if not the default scaling params option will be used, with default
    configuration options. The model can be individually specified.

    Returns the reflection table, with added columns 'inverse_scale_factor' and
    'inverse_scale_factor_variance'."""

    if not params:
        phil_scope = phil.parse(
            """
      include scope dials.algorithms.scaling.model.model.model_phil_scope
      include scope dials.algorithms.scaling.scaling_options.phil_scope
      include scope dials.algorithms.scaling.scaling_refiner.scaling_refinery_phil_scope
    """,
            process_includes=True,
        )
        parser = ArgumentParser(phil=phil_scope, check_format=False)
        params, _ = parser.parse_args(args=[], quick_parse=True)

    params.model = model

    from dials.algorithms.scaling.scaler_factory import SingleScalerFactory

    experiments = create_scaling_model(params, experiment, [reflection_table])
    scaler = SingleScalerFactory.create(params, experiments[0], reflection_table)
    from dials.algorithms.scaling.algorithm import scaling_algorithm

    scaler = scaling_algorithm(scaler)
    return scaler.reflection_table


def create_scaling_model(params, experiments, reflections):
    """Loop through the experiments, creating the scaling models."""
    autos = [None, Auto, "auto", "Auto"]
    use_auto_model = params.model in autos
    # Determine non-auto model to use outside the loop over datasets.
    if not use_auto_model:
        model_class = None
        for entry_point in pkg_resources.iter_entry_points("dxtbx.scaling_model_ext"):
            if entry_point.name == params.model:
                model_class = entry_point.load()
                break
        if not model_class:
            raise ValueError(f"Unable to create scaling model of type {params.model}")

    for expt, refl in zip(experiments, reflections):
        if not expt.scaling_model or params.overwrite_existing_models:
            # need to make a new model
            if use_auto_model:
                if not expt.scan:
                    model = KBScalingModel
                else:  # set model as physical unless scan < 1.0 degree
                    osc_range = expt.scan.get_oscillation_range()
                    abs_osc_range = abs(osc_range[1] - osc_range[0])
                    if abs_osc_range < 1.0:
                        model = KBScalingModel
                    else:
                        model = PhysicalScalingModel
            else:
                model = model_class
            expt.scaling_model = model.from_data(params, expt, refl)
        else:
            # allow for updating of an existing model.
            expt.scaling_model.update(params)
    return experiments


def create_Ih_table(
    experiments, reflections, selections=None, n_blocks=1, anomalous=False
):
    """Create an Ih table from a list of experiments and reflections. Optionally,
    a selection list can also be given, to select data from each reflection table.
    Allow an unequal number of experiments and reflections, as only need to
    extract one space group value (can optionally check all same if many)."""
    if selections:
        assert len(selections) == len(
            reflections
        ), """Must have an equal number of
    reflection tables and selections in the input lists."""
    space_group_0 = experiments[0].crystal.get_space_group()
    for experiment in experiments:
        assert (
            experiment.crystal.get_space_group() == space_group_0
        ), """The space
    groups of all experiments must be equal."""
    input_tables = []
    indices_lists = []
    for i, reflection in enumerate(reflections):
        if "inverse_scale_factor" not in reflection:
            reflection["inverse_scale_factor"] = flex.double(reflection.size(), 1.0)
        if selections:
            input_tables.append(reflection.select(selections[i]))
            indices_lists.append(selections[i].iselection())
        else:
            input_tables.append(reflection)
            indices_lists = None
    Ih_table = IhTable(
        input_tables,
        space_group_0,
        indices_lists,
        nblocks=n_blocks,
        anomalous=anomalous,
    )
    return Ih_table


def scaled_data_as_miller_array(
    reflection_table_list,
    experiments,
    best_unit_cell=None,
    anomalous_flag=False,
    wavelength=None,
):
    """Get a scaled miller array from an experiment and reflection table."""
    if len(reflection_table_list) > 1:
        joint_table = flex.reflection_table()
        for reflection_table in reflection_table_list:
            # better to just create many miller arrays and join them?
            refl_for_joint_table = flex.reflection_table()
            for col in [
                "miller_index",
                "intensity.scale.value",
                "inverse_scale_factor",
                "intensity.scale.variance",
            ]:
                refl_for_joint_table[col] = reflection_table[col]
            good_refl_sel = ~reflection_table.get_flags(
                reflection_table.flags.bad_for_scaling, all=False
            )
            refl_for_joint_table = refl_for_joint_table.select(good_refl_sel)
            joint_table.extend(refl_for_joint_table)
    else:
        reflection_table = reflection_table_list[0]
        good_refl_sel = ~reflection_table.get_flags(
            reflection_table.flags.bad_for_scaling, all=False
        )
        joint_table = reflection_table.select(good_refl_sel)
    # Filter out negative scale factors to avoid merging statistics errors.
    # These are not removed from the output data, as it is likely one would
    # want to do further analysis e.g. delta cc1/2 and rescaling, to exclude
    # certain data and get better scale factors for all reflections.
    pos_scales = joint_table["inverse_scale_factor"] > 0
    if pos_scales.count(False) > 0:
        logger.info(
            """There are %s reflections with non-positive scale factors which
will not be used for calculating merging statistics""",
            pos_scales.count(False),
        )
        joint_table = joint_table.select(pos_scales)

    if best_unit_cell is None:
        best_unit_cell = determine_best_unit_cell(experiments)
    miller_set = miller.set(
        crystal_symmetry=crystal.symmetry(
            unit_cell=best_unit_cell,
            space_group=experiments[0].crystal.get_space_group(),
            assert_is_compatible_unit_cell=False,
        ),
        indices=joint_table["miller_index"],
        anomalous_flag=anomalous_flag,
    )
    i_obs = miller.array(
        miller_set,
        data=joint_table["intensity.scale.value"] / joint_table["inverse_scale_factor"],
    )
    i_obs.set_observation_type_xray_intensity()
    i_obs.set_sigmas(
        flex.sqrt(joint_table["intensity.scale.variance"])
        / joint_table["inverse_scale_factor"]
    )
    if not wavelength:
        wavelength = np.mean([expt.beam.get_wavelength() for expt in experiments])
    i_obs.set_info(
        miller.array_info(
            source="DIALS",
            source_type="reflection_tables",
            wavelength=wavelength,
        )
    )
    return i_obs


def determine_best_unit_cell(experiments):
    """Set the median unit cell as the best cell, for consistent d-values across
    experiments."""
    uc_params = [flex.double() for i in range(6)]
    for exp in experiments:
        unit_cell = (
            exp.crystal.get_recalculated_unit_cell() or exp.crystal.get_unit_cell()
        )
        for i, p in enumerate(unit_cell.parameters()):
            uc_params[i].append(p)
    best_unit_cell = uctbx.unit_cell(parameters=[flex.median(p) for p in uc_params])
    if len(experiments) > 1:
        logger.info("Using median unit cell across experiments : %s", best_unit_cell)
    return best_unit_cell


def merging_stats_from_scaled_array(
    scaled_miller_array, n_bins=20, use_internal_variance=False, anomalous=True
):
    """Calculate the normal and anomalous merging statistics."""

    if scaled_miller_array.is_unique_set_under_symmetry():
        raise DialsMergingStatisticsError(
            "Dataset contains no equivalent reflections, merging statistics "
            "cannot be calculated."
        )
    try:
        result = iotbx.merging_statistics.dataset_statistics(
            i_obs=scaled_miller_array,
            n_bins=n_bins,
            anomalous=False,
            sigma_filtering=None,
            eliminate_sys_absent=False,
            use_internal_variance=use_internal_variance,
            cc_one_half_significance_level=0.01,
        )
    except (RuntimeError, Sorry) as e:
        raise DialsMergingStatisticsError(
            f"Error encountered during merging statistics calculation:\n{e}"
        )

    anom_result = None

    if anomalous:
        intensities_anom = scaled_miller_array.as_anomalous_array()
        intensities_anom = intensities_anom.map_to_asu().customized_copy(
            info=scaled_miller_array.info()
        )
        if intensities_anom.is_unique_set_under_symmetry():
            logger.warning(
                "Anomalous dataset contains no equivalent reflections, anomalous "
                "merging statistics cannot be calculated."
            )
        else:
            try:
                anom_result = iotbx.merging_statistics.dataset_statistics(
                    i_obs=intensities_anom,
                    n_bins=n_bins,
                    anomalous=True,
                    sigma_filtering=None,
                    cc_one_half_significance_level=0.01,
                    eliminate_sys_absent=False,
                    use_internal_variance=use_internal_variance,
                )
            except (RuntimeError, Sorry) as e:
                logger.warning(
                    "Error encountered during anomalous merging statistics "
                    "calculation:\n%s",
                    e,
                    exc_info=True,
                )

    return result, anom_result


def create_datastructures_for_reference_file(
    experiments, reference_file, anomalous=True, d_min=2.0
):
    # If the file is a model file, then d_min is used to determine the highest
    # resolution calculated intensities.
    wavelength = np.mean([expt.beam.get_wavelength() for expt in experiments])
    intensities = intensities_from_reference_file(reference_file, d_min, wavelength)
    if not anomalous:
        intensities = intensities.as_non_anomalous_array().merge_equivalents().array()

    table = flex.reflection_table()
    table["intensity"] = intensities.data()
    table["miller_index"] = intensities.indices()
    table["id"] = flex.int(table.size(), len(experiments))
    table["d"] = intensities.d_spacings().data()

    if intensities.sigmas():  # only the case for a data file
        table["variance"] = intensities.sigmas() ** 2
        logger.info(f"Extracted {table.size()} intensities from reference file")
        table = table.select(table["variance"] > 0.0)
        if table.size() == 0:
            raise ValueError(
                "No reflections with positive sigma remain after filtering"
            )

    table.set_flags(flex.bool(table.size(), False), table.flags.bad_for_scaling)
    table.set_flags(flex.bool(table.size(), True), table.flags.integrated)

    expt = Experiment()
    expt.crystal = deepcopy(experiments[0].crystal)
    params = Mock()
    params.KB.decay_correction.return_value = False
    expt.scaling_model = KBScalingModel.from_data(params, [], [])
    expt.scaling_model.set_scaling_model_as_scaled()  # Set as scaled to fix scale.
    expt.identifier = ersatz_uuid4()

    table.experiment_identifiers()[len(experiments)] = expt.identifier

    table.experiment_identifiers()[len(experiments)] = expt.identifier

    return expt, table


def create_datastructures_for_target_mtz(experiments, mtz_file, anomalous=True):
    """
    Read a merged mtz file and extract miller indices, intensities and variances.
    Deprecated, retained for backwards compability.
    """
    return create_datastructures_for_reference_file(experiments, mtz_file, anomalous)


def create_datastructures_for_structural_model(
    experiments, model_file, anomalous=True, d_min=2.0
):
    """Read a cif/pdb file, calculate intensities. Return an experiment and
    reflection table to be used for the structural model in scaling."""
    return create_datastructures_for_reference_file(
        experiments, model_file, anomalous, d_min
    )
